/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for handling constantpool item */

#include <string.h>	/* for strlen(), strcmp() */

#include <native.h>	/* for native methods */
#include <sys_api.h>	/* for sys*() */
#include <utf.h>	/* for unicode2utf() */

#include "NET_shudo_moba_reflect_ConstantPool.h"

#include "moba.h"


#if (JDK_VER < 110)
#define clazzConstantPool(clazz) cbConstantPool(unhand(clazz))
#else
#define clazzConstantPool(clazz) cbConstantPool(clazz)
#endif


/*
 * constant pool count ֤
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_count(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz) {
  ClassClass *cb;
#if (JDK_VER < 110)
  cb = unhand(clazz);
#else
  cb = clazz;
#endif

  return (int32_t)cbConstantPoolCount(cb);
}


/*
 * constantpool μ (CONSTANT_*) ֤
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_type(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t index) {
  cp_item_type *constant_pool = clazzConstantPool(clazz);
  unsigned char *typeTable;
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif
  return (int32_t)typeTable[index] & CONSTANT_POOL_ENTRY_TYPEMASK;
}


/*
 * (String)Ȥ Utf8 ͤ롣
 */
extern struct Hjava_lang_String *NET_shudo_moba_reflect_ConstantPool_getUtf8Value(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz ,int32_t index) {
  cp_item_type *constant_pool = clazzConstantPool(clazz);
  char *utf = constant_pool[index].cp;
  return makeJavaString(utf, strlen(utf));
}

/*
 * (int)ͤ롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_getIntValue(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t index) {
  cp_item_type *constant_pool = clazzConstantPool(clazz);
  return (int32_t)(constant_pool[index].i);
}

/*
 * (float)ͤ롣
 */
extern float NET_shudo_moba_reflect_ConstantPool_getFloatValue(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t index) {
  cp_item_type *constant_pool = clazzConstantPool(clazz);
  return constant_pool[index].f;
}

/*
 * (long)ͤ롣
 */
extern int64_t NET_shudo_moba_reflect_ConstantPool_getLongValue(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t index) {
  cp_item_type *constant_pool = clazzConstantPool(clazz);
  Java8 t1;

  t1.l = GET_INT64(t1, &constant_pool[index]);
  return t1.l;
}

/*
 * (double)ͤ롣
 */
extern double NET_shudo_moba_reflect_ConstantPool_getDoubleValue(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t index) {
  cp_item_type *constant_pool = clazzConstantPool(clazz);
  Java8 t1;

  t1.d = GET_DOUBLE(t1, &constant_pool[index]);
  return t1.d;
}

/*
 * (Class)ͤ롣
 */
extern struct Hjava_lang_Class *NET_shudo_moba_reflect_ConstantPool_getClassValue(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t index) {
  ClassClass *cb = NULL;
  ClassClass *target = NULL;
  cp_item_type *constant_pool = clazzConstantPool(clazz);

  unsigned char *typeTable;
  int resolved;

#if (JDK_VER < 110)
  cb = unhand(clazz);
  typeTable =
	(unsigned char *)constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  cb = clazz;
  typeTable = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif
  resolved = typeTable[index] & CONSTANT_POOL_ENTRY_RESOLVED;

  if (resolved) {
#if (JDK_VER < 110)
    target = (ClassClass *)constant_pool[index].cp;
#else
    target = constant_pool[index].clazz;
#endif
  }
  else {
    char *name = constant_pool[constant_pool[index].i].cp;
    target = FindClassFromClass(EE(), name, TRUE, cb);
    if (!target) {
      /* not reached */
      printf("class %s is not found.\n", target);  fflush(stdout);
      return 0;
    }
  }

  return cbHandle(target);
}

/*
 * (String)ͤ롣
 */
struct Hjava_lang_String;
extern struct Hjava_lang_String *NET_shudo_moba_reflect_ConstantPool_getStringValue(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t index) {
  cp_item_type *constant_pool = clazzConstantPool(clazz);

  unsigned char *typeTable;
  int resolved;

#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif
  resolved = typeTable[index] & CONSTANT_POOL_ENTRY_RESOLVED;

  if (resolved) {
#if (JDK_VER < 110)
    return (struct Hjava_lang_String *)constant_pool[index].cp;
#else
    return constant_pool[index].str;
#endif
  }
  else {
    int utfIndex = constant_pool[index].i;
    char *utf = constant_pool[constant_pool[index].i].cp;
    return makeJavaString(utf, strlen(utf));
  }
}


/*
 * ꤷ int ͤؤǤΥǥå롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_intIndex(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int32_t value) {
  ClassClass *classCb;
#if (JDK_VER < 110)
  classCb = unhand(clazz);
#else
  classCb = clazz;
#endif

  return searchNBytes(classCb, CONSTANT_Integer, &value, sizeof(int32_t));
}

/*
 * ꤷ float ͤؤǤΥǥå롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_floatIndex(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, float value) {
  ClassClass *classCb;
#if (JDK_VER < 110)
  classCb = unhand(clazz);
#else
  classCb = clazz;
#endif

  return searchNBytes(classCb, CONSTANT_Float, &value, sizeof(float));
}

/*
 * ꤷ long ͤؤǤΥǥå롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_longIndex(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, int64_t value) {
  ClassClass *classCb;
#if (JDK_VER < 110)
  classCb = unhand(clazz);
#else
  classCb = clazz;
#endif

  return searchNBytes(classCb, CONSTANT_Long, &value, sizeof(int64_t));
}

/*
 * ꤷ double ͤؤǤΥǥå롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_doubleIndex(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, double value) {
  ClassClass *classCb;
#if (JDK_VER < 110)
  classCb = unhand(clazz);
#else
  classCb = clazz;
#endif

  return searchNBytes(classCb, CONSTANT_Double, &value, sizeof(double));
}

/*
 * ꤷ boolean ͤؤǤΥǥå롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_booleanIndex(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, /*boolean*/ int32_t value) {
  ClassClass *classCb;
  int32_t intFalse = 0, intTrue = 1;
#if (JDK_VER < 110)
  classCb = unhand(clazz);
#else
  classCb = clazz;
#endif

  if (value) {
#if 1	/* (int)1 Ǥ CP õ */
    return searchNBytes(classCb, CONSTANT_Integer,
					&intTrue, sizeof(int32_t));
#else	/* 0 Ǥʤ (int)  CP õ */
    return searchNBytesNotEqual(classCb, CONSTANT_Integer,
					&intFalse, sizeof(int32_t));
#endif
  }
  else {
    return searchNBytes(classCb, CONSTANT_Integer, &intFalse, sizeof(int32_t));
  }
}

/*
 * ꤷ UTF8 ʸؤǤΥǥå롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_utf8Index(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, struct Hjava_lang_String *tgtString) {
  ClassClass *classCb;
  HArrayOfChar *value;
  unicode *body;
  char *utfString;
  int index;

#if (JDK_VER < 110)
  classCb = unhand(clazz);
#else
  classCb = clazz;
#endif

  value = unhand(tgtString)->value;	/* String -> char[] */
  body = unhand(value)->body;		/* char[] -> unicode[] */

  /* UCS2 -> UTF8 Ѵ */
  utfString = unicode2utf(body, obj_length(value), (char *)NULL, 0);
  index = searchUtf8(classCb, utfString);
  sysFree(utfString);

  return index;
}

/*
 * ꤷ String ؤǤΥǥå롣
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_stringIndex(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, struct Hjava_lang_String *tgtString) {
  ClassClass *classCb;
  cp_item_type *constantPool;
  unsigned char *typeTable;
  int itemCount;

  int ret;
  int i;

  int tgtStrLen;
  HArrayOfChar *tgtValue;
  unicode *tgtBody;
  char *tgtBodyInUtf;
  int utfIndex, strIndex;

#if (JDK_VER < 110)
  classCb = unhand(clazz);
#else
  classCb = clazz;
#endif
  constantPool = cbConstantPool(classCb);
#if (JDK_VER < 110)
  typeTable = (unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif
  itemCount = cbConstantPoolCount(classCb);


  tgtStrLen = javaStringLength(tgtString);
  tgtValue = unhand(tgtString)->value;	/* String -> char[] */
  tgtBody = unhand(tgtValue)->body;	/* char[] -> (unicode *) */

  /* оݤ UCS2 -> UTF8 Ѵ */
  tgtBodyInUtf = unicode2utf(tgtBody, obj_length(tgtValue), (char *)NULL, 0);
  utfIndex = searchUtf8(classCb, tgtBodyInUtf);

  if (utfIndex < 0) {  ret = -1;  goto returnStringIndex;  }

  for (i = 1; i < itemCount; i++) {
    unsigned char type = typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK;
	/* type is CONSTANT_* */

    if ((type == CONSTANT_String)) {
      if (typeTable[i] & CONSTANT_POOL_ENTRY_RESOLVED) {
	Hjava_lang_String *itemString;
	unicode *itemBody;
#if (JDK_VER < 110)
	itemString = (Hjava_lang_String *)constantPool[i].cp;
#else
	itemString = constantPool[i].str;
#endif
	itemBody = unhand(unhand(itemString)->value)->body;
	if (!memcmp(tgtBody, itemBody, sizeof(unicode) * tgtStrLen)) {
	  ret = i;  goto returnStringIndex;
	}
      }
      else {
	char *itemBody = constantPool[constantPool[i].i].cp;
	int itemBodyLen = strlen(itemBody);
	if (!memcmp(tgtBodyInUtf, itemBody, itemBodyLen)) {
	  ret = i;  goto returnStringIndex;
	}
      }	/* CP entry  resolve Ƥ뤫 */
    }	/* CONSTANT_String  */
  }

returnStringIndex:
  sysFree(tgtBodyInUtf);

#ifdef DEBUG
  if (ret <= 0) {
    printf("ConstantPool#stringIndex(class `%s', string `%s'):"
	 " entry is not found in CP.\n", cbName(classCb), tgtBodyInUtf);
    fflush(stdout);
  }
#endif
  return ret;
}

/*
 * ꤷ饹ؤǤΥǥå롣
 * null ꤵ줿 0 ֤(java.lang.Object  superclass)
 */
extern int32_t NET_shudo_moba_reflect_ConstantPool_classIndex(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz, struct Hjava_lang_Class *target) {
  ClassClass *classCb;
  ClassClass *targetCb;

  if (!clazz || !target)  return 0;

#if (JDK_VER < 110)
  classCb = unhand(clazz);
  targetCb = unhand(target);
#else
  classCb = clazz;
  targetCb = target;
#endif

  return searchClazz(classCb, targetCb);
}


/*
 * classfile ΥեޥåȤ constant pool Ϥ롣
 */
extern void NET_shudo_moba_reflect_ConstantPool_writeInClassfile(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_io_OutputStream *out, struct Hjava_lang_Class *clazz) {
  ClassClass *cb;
  int cpCount;
  cp_item_type *constantPool;
  unsigned char *typeTable;
  int i;

#if (JDK_VER < 110)
  cb = unhand(clazz);
#else
  cb = clazz;
#endif
  cpCount = cbConstantPoolCount(cb);
  constantPool = cbConstantPool(cb);
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif

#ifdef DEBUG
  printf("MetaClass#writeCPInClassfile() called.\n");  fflush(stdout);
#endif

  for (i = 1; i < cpCount; i++) {
    unsigned char type = typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK;
    int resolved = typeTable[i] & CONSTANT_POOL_ENTRY_RESOLVED;
	/* type is CONSTANT_* */
    Java8 t1;	/* 64bit ͳǼ */

    uint16_t len, index0, index1;
#ifdef DEBUG
    printf("CP[%d]: %s (%d)\n", i, cpTypeString(type), type);
    fflush(stdout);
#endif

    if (type == 0)  continue;	/* ľͤ 64bit ä */


    /* tag: CONSTANT_* */
    if ((writeChar(out, (int)type)) < 0)  goto exc_occurred;

    switch (type) {
    case CONSTANT_Utf8:
    case CONSTANT_Unicode:
      len = strlen(constantPool[i].cp);
      if ((writeInt16(out, (int16_t *)&len)) < 0)  goto exc_occurred;
      if ((writeBytes(out, constantPool[i].cp, (int)len)) < 0)
	goto exc_occurred;
	/* _Utf8: VM Ǥ UTF8 ΤޤޤʤΤǤΤޤ޽ */
      break;
    case CONSTANT_Integer:
    case CONSTANT_Float:
      if ((writeInt32(out, (int32_t *)&constantPool[i])) < 0)
	goto exc_occurred;
      break;
    case CONSTANT_Long:
    case CONSTANT_Double:
      if ((writeInt64(out, (int64_t *)&constantPool[i])) < 0)
	goto exc_occurred;
      break;
    case CONSTANT_Class:
      if (resolved) {
	ClassClass *clazz;
#if (JDK_VER < 110)
	clazz = (ClassClass *)constantPool[i].cp;
#else
	clazz = constantPool[i].clazz;
#endif
	index0 = searchUtf8(cb, cbName(clazz));
      }
      else
	index0 = constantPool[i].i;
      if ((writeInt16(out, (int16_t *)&index0)) < 0)  goto exc_occurred;
      break;
    case CONSTANT_String:
      if (resolved) {
	Hjava_lang_String *hStr;
	HArrayOfChar *value;
	unicode *body;
	char *buf;

#ifdef DEBUG
	printf("constantPool[%d]: 0x%x.\n", i, *((int *)&constantPool[i]));
	fflush(stdout);
#endif
#if (JDK_VER < 110)
	hStr = (Hjava_lang_String *)constantPool[i].cp;
#else
	hStr = constantPool[i].str;
#endif /* JDK_VER < 110 */
	value = unhand(hStr)->value;
#ifdef DEBUG
	printf("value: 0x%x.\n", (int)value);  fflush(stdout);
#endif
	body = unhand(value)->body;
#ifdef DEBUG
	printf("body: 0x%x.\n", (int)body);  fflush(stdout);
#endif

	/* UCS2 -> UTF8 Ѵ */
	buf = unicode2utf(body, obj_length(value), (char *)NULL, 0);
	index0 = searchUtf8(cb, buf);
	sysFree(buf);
      }
      else
	index0 = constantPool[i].i;
      if ((writeInt16(out, (int16_t *)&index0)) < 0)  goto exc_occurred;
      break;
    case CONSTANT_Fieldref:
      if (resolved) {
	struct fieldblock *fb;
#if (JDK_VER < 110)
	fb = (struct fieldblock *)constantPool[i].cp;
#else
	fb = constantPool[i].fb;
#endif /* JDK_VER < 110 */
	index0 = searchClazz(cb, fb->clazz);
	index1 = searchNameAndTypeByField(cb, fb);
	if ((writeInt16(out, (int16_t *)&index0)) < 0)  goto exc_occurred;
	if ((writeInt16(out, (int16_t *)&index1)) < 0)  goto exc_occurred;
      }
      else
	if ((writeInt32(out, (int32_t *)&(constantPool[i].i))) < 0)
	  goto exc_occurred;
      break;
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
      if (resolved) {
	struct methodblock *mb;
#if (JDK_VER < 110)
	mb = (struct methodblock *)constantPool[i].cp;
#else
	mb = constantPool[i].mb;
#endif /* JDK_VER < 110 */
	index0 = searchClazz(cb, mb->fb.clazz);
	index1 = searchNameAndTypeByMethod(cb, mb);
#ifdef DEBUG
	printf("Class index      : %d\n", (int)index0);
	printf("NameAndType index: %d\n", (int)index1);
	fflush(stdout);
#endif
	if ((writeInt16(out, (int16_t *)&index0)) < 0)  goto exc_occurred;
	if ((writeInt16(out, (int16_t *)&index1)) < 0)  goto exc_occurred;
      }
      else
	if ((writeInt32(out, (int32_t *)&(constantPool[i].i))) < 0)  goto exc_occurred;
      break;
    case CONSTANT_NameAndType:
      if (resolved) {
	struct fieldblock *field;
	struct methodblock *method;
	uint32_t id = constantPool[i].i;

#ifdef DEBUG
	printf("ID of NameAndType: %d\n", id);
	fflush(stdout);
#endif
	if (field = findFieldByID(cb, id)) {
		/* 饹⤫ field õ */
	  index0 = searchUtf8(cb, field->name);
	  index1 = searchUtf8(cb, field->signature);
	}
	else if (method = findMethodByID(cb, id)) {
		/* 饹⤫ method õ */
	  index0 = searchUtf8(cb, method->fb.name);
	  index1 = searchUtf8(cb, method->fb.signature);
	}
	else if (field = findFieldInVM(id)) {
		/* JavaVM ⤫ field õ */
	  index0 = searchUtf8(cb, field->name);
	  index1 = searchUtf8(cb, field->signature);
	}
	else if (method = findMethodInVM(id)) {
		/* JavaVM ⤫ method õ */
	  index0 = searchUtf8(cb, method->fb.name);
	  index1 = searchUtf8(cb, method->fb.signature);
	}
	else {
#ifdef DEBUG
	  printf("Any field or method has ID(%d) is not found.\n", id);
	  fflush(stdout);
#endif
	  index0 = index1 = -1;
	}

	if ((writeInt16(out, (int16_t *)&index0)) < 0)  goto exc_occurred;
	if ((writeInt16(out, (int16_t *)&index1)) < 0)  goto exc_occurred;
      }
      else
	if ((writeInt32(out, (int32_t *)&(constantPool[i].i))) < 0)
	  goto exc_occurred;
      break;
    default:
      /* not reached. */
      /* invalid tag */
      break;
    }
  }

  return;

exc_occurred:
  THROW_IOEXCEPTION();
}


/*
 * classfile  constant pool Хȿ롣
 */
extern int64_t NET_shudo_moba_reflect_ConstantPool_lengthInClassfile(struct HNET_shudo_moba_reflect_ConstantPool *_this, struct Hjava_lang_Class *clazz) {
  ClassClass *cb;
  int cpCount;
  cp_item_type *constantPool;
  unsigned char *typeTable;
  int64_t result = 0;
  int i;

#if (JDK_VER < 110)
  cb = unhand(clazz);
#else
  cb = clazz;
#endif
  cpCount = cbConstantPoolCount(cb);
  constantPool = cbConstantPool(cb);
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif

  for (i = 1; i < cpCount; i++) {
    unsigned char type = typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK;

    switch (type) {
    case CONSTANT_Utf8:
    case CONSTANT_Unicode:
      result += 3;
      result += strlen(constantPool[i].cp);
      break;
    case CONSTANT_Integer:
    case CONSTANT_Float:
      result += 5;
      break;
    case CONSTANT_Long:
    case CONSTANT_Double:
      result += 9;
      break;
    case CONSTANT_Class:
    case CONSTANT_String:
      result += 3;
      break;
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_NameAndType:
      result += 5;
      break;
    default:
      break;
    }
  }

  return result;
}


/*
 * Global Functions
 */

/*
 * CONSTANT_* бʸ֤
 */
char *cpTypeString(int type) {
  static char *typeStr[] = {
	"(null)", "Utf8", "Unicode", "Integer", "Float",
	"Long", "Double", "Class", "String",
	"Fieldref", "Methodref", "InterfaceMethodref",
	"NameAndType"
  };

  if ((type < CONSTANT_Utf8) || (type > CONSTANT_NameAndType)) {
    return "(null)";
  }

  return typeStr[type];
}


/*
 * CONSTANT_Utf8 Υȥõǥå֤
 * Ĥʤ -1 ֤
 */
int searchUtf8(ClassClass *cb, char *utfStr) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable;
  int i;
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif

  for (i = 1; i < itemCount; i++) {
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == CONSTANT_Utf8) {
      if (!strcmp(constantPool[i].cp, utfStr))
	return i;
    }
  }

#ifdef DEBUG
  printf("searchUtf8(class `%s', utfStr `%s'):"
	 " entry is not found in CP.\n", cbName(cb), utfStr);
  fflush(stdout);
#endif
  return -1;
}

/*
 * CONSTANT_Class Υȥõǥå֤
 * Ĥʤ -1 ֤
 */
int searchClazz(ClassClass *cb, ClassClass *targetCb) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable;
  int utfIndex;
  int i;
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif

  if ((utfIndex = searchUtf8(cb, cbName(targetCb))) < 0)
    goto searchClazzErrorReturn;

  for (i = 1; i < itemCount; i++) {
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == CONSTANT_Class) {
      if (typeTable[i] & CONSTANT_POOL_ENTRY_RESOLVED) {
	ClassClass *clazz;
#if (JDK_VER < 110)
	clazz = (ClassClass *)constantPool[i].cp;
#else
	clazz = constantPool[i].clazz;
#endif
#ifdef COMPARE_CLASS_BY_NAME	/* 饹̾ */
	if (!strcmp(cbName(clazz), cbName(targetCb)))
#else	/* clazz Υɥ쥹 */
	if (clazz == targetCb)
#endif
	  return i;
      }
      else {
	if (!strcmp(constantPool[constantPool[i].i].cp, cbName(targetCb)))
	  return i;
      }
    }
  }

searchClazzErrorReturn:
#ifdef DEBUG
  printf("searchClazz(): class `%s' is not found in CP.\n", cbName(targetCb));
  fflush(stdout);
#endif
  return -1;
}


/*
 * (struct fieldblock *) CONSTANT_NameAndType Υȥõ
 * ǥå֤
 * Ĥʤ -1 ֤
 */
int searchNameAndTypeByField(ClassClass *cb, struct fieldblock *fb) {
  uint32_t id = NameAndTypeToHash(fb->name, fb->signature);
  return searchNameAndType(cb, id);
}


/*
 * (struct methodblock *) CONSTANT_NameAndType Υȥõ
 * ǥå֤
 * Ĥʤ -1 ֤
 */
int searchNameAndTypeByMethod(ClassClass *cb, struct methodblock *mb) {
  uint32_t id = NameAndTypeToHash(mb->fb.name, mb->fb.signature);
  return searchNameAndType(cb, id);
}


/*
 * NameAndTypeToHash()ǵޤ ID  CONSTANT_NameAndType Υȥõ
 * ǥå֤
 * Ĥʤ -1 ֤
 */
int searchNameAndType(ClassClass *cb, uint32_t id) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable;
  int i;
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif

  for (i = 1; i < itemCount; i++) {
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) ==
		CONSTANT_NameAndType) {
      if (constantPool[i].i == id)
	return i;
    }
  }

#ifdef DEBUG
  printf("searchNameAndType(class `%s', id %d):"
	 " entry is not found in CP.\n", cbName(cb), id);
  fflush(stdout);
#endif
  return -1;
}


/*
 * Local Functions
 */

/*
 * ȥõǥå֤
 * ߤĤʤ -1 ֤
 */
int searchNBytes(ClassClass *cb,
	int type/* CONSTANT_* */, char *data, int size) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable;
  int i;
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif

  for (i = 1; i < itemCount; i++) {
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == type) {
      if (!memcmp(constantPool + i, data, size))
	return i;
    }
  }

#ifdef DEBUG
  printf("searchNBytes(class `%s', type %d, data 0x%x, size %d):"
	 " entry is not found in CP.\n", cbName(cb), type, (int)data, size);
  fflush(stdout);
#endif
  return -1;
}

/*
 * ͤפʤȥõǥå֤
 * ߤĤʤ -1 ֤
 */
int searchNBytesNotEqual(ClassClass *cb,
	int type/* CONSTANT_* */, char *data, int size) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable;
  int i;
#if (JDK_VER < 110)
  typeTable =
	(unsigned char *)constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
#else
  typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
#endif

  for (i = 1; i < itemCount; i++) {
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == type) {
      if (memcmp(constantPool + i, data, size))
	return i;
    }
  }

#ifdef DEBUG
  printf("searchNBytesNotEqual(class `%s', type %d, data 0x%x, size %d):"
	 " entry is not found in CP.\n", cbName(cb), type, (int)data, size);
  fflush(stdout);
#endif
  return -1;
}
