/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for MetaClass */

#include <stdio.h>	/* for NULL */
#include <string.h>	/* for strlen() */

#include <native.h>	/* for native methods */
#include <utf.h>	/* for unicode2utf() */

#include "java_lang_ClassLoader.h"	/* for ClassLoader_defineClass0() */
#include "NET_shudo_moba_reflect_MetaClass.h"

#include "moba.h"


/*
 * signature  Class ֥Ȥ롣
 */
extern struct Hjava_lang_Class *NET_shudo_moba_reflect_MetaClass_classBySignature(struct HNET_shudo_moba_reflect_MetaClass *_this, struct Hjava_lang_String *sigStr, struct Hjava_lang_ClassLoader *loader) {
  ExecEnv *ee = EE();
  char *sig = makeCString(sigStr);
  ClassClass *cb;

  /* Lclass_name;  class_name  */
  if (sig[0] == SIGNATURE_CLASS /* 'L' */) {
    char *ptr = ++sig;
    while ((*ptr != SIGNATURE_ENDCLASS /* `;' */) && (*ptr != '\0'))  ptr++;
    *ptr = '\0';
  }

  if ((sig[0] != SIGNATURE_ARRAY) && loader)
    cb = ClassLoaderFindClass(ee, loader, sig, TRUE);
  else
    cb = FindClass(ee, sig, TRUE);

  if (!cb) {
    if (!exceptionOccurred(ee))
      SignalError(0, "java.lang.ClassNotFoundException", sig);
    return 0;
  }
  return cbHandle(cb);
}


/*
 * Υ饹 signature 롣
 */
extern struct Hjava_lang_String *NET_shudo_moba_reflect_MetaClass_signature(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  char *name = cbName(cb);
  return makeJavaString(name, strlen(name));
}


/*
 * ֥ () Υ󥹥󥹤롣
 * ˤĤƤ MetaClass#newArray() Ѥ롣
 */
extern struct Hjava_lang_Object *NET_shudo_moba_reflect_MetaClass_newInstance(struct HNET_shudo_moba_reflect_MetaClass *_this, struct Hjava_lang_Class *aClass) {
  ExecEnv *ee = EE();
  ClassClass *cb, *cbtmp;
  HObject *obj;

  uint32_t initID;
  struct methodblock *initMethod;
  int i;

  if (!aClass)  return NULL;

#if (JDK_VER < 110)
  cb = (ClassClass *)unhand(aClass);
#else
  cb = (ClassClass *)aClass;
#endif
  if (!cb)  return NULL;


  /* The code below is derived from runtime/interpretter.c */

  /* search constructor */
  /* java.lang.Object ޤ superclass éʤΥ󥹥ȥ饯õ */
  initID = NameAndTypeToHash("<init>", "()V");
  cbtmp = cb;
  while (cbtmp) {
    for (i = cbMethodsCount(cbtmp), initMethod = cbMethods(cbtmp); --i >= 0;
	 initMethod++) {
      if (initMethod->fb.ID == initID)  break;
    }
    if (i >= 0)	/* constructor without args. is not found. */
      break;
    else
      cbtmp = cbSuperclass(cbtmp);
  }

  /* create an instance */
  if ((obj = newobject(cb, 0, ee)) == 0) {
    /* 󥹥ȥ饯ĤʤƤ instance ϺäƤ */
#ifdef DEBUG
    printf("MetaClass#newInstance(): newobject() failed.\n");
    fflush(stdout);
#endif
    SignalError(0, JAVAPKG "OutOfMemoryError", 0);
    return 0;
  }

  /* check if the class has constructor without args. */
  if (!cbtmp) {
    /* not reached. */
#ifdef DEBUG
    printf("MetaClass#newInstance(): constructor is not found.\n");
    fflush(stdout);
#endif
    SignalError(0, "java/lang/NoSuchMethodError", 0);
  }
  else {
    /* call the constructor */
    if (java_monitor) {
      struct methodblock *caller;
      long start_time, end_time;
      start_time = now();
      do_execute_java_method(ee, obj,
		0, 0, initMethod, 0, /* argument */ NULL);
      end_time = now();
      caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
      if (!caller)  caller = ee->current_frame->current_method;
      java_mon(caller, initMethod, end_time-start_time);
    }
    else {
      do_execute_java_method(ee, obj,
		0, 0, initMethod, 0, /* argument */ NULL);
    }
    /* if (exceptionOccurred(ee))  return 0; */
  }

#ifdef DEBUG
  printf("MetaClass#newInstance(%s) returns 0x%x\n", cbName(cb), (int)obj);
  printf(" class of obj: %s\n", cbName(obj_classblock(obj)));
  fflush(stdout);

  if (exceptionOccurred(ee)) {
    printf("Exception occurred in MetaClass#newInstance().\n");
    fflush(stdout);

    exceptionDescribe(ee);
  }

  fflush(stdout);
#endif
  return obj;
}


/*
 * Υ󥹥󥹤롣
 * ֥ȤˤĤƤ MetaClass#newInstance() Ѥ롣
 */
extern struct Hjava_lang_Object *NET_shudo_moba_reflect_MetaClass_newArray(struct HNET_shudo_moba_reflect_MetaClass *_this, struct Hjava_lang_Class *hclass, int32_t len) {
  Hjava_lang_Object *obj = NULL;	/*  */
  ClassClass *arrayClazz, *elemClazz = NULL;
  char *sig;  int sigLength;

  /*  class, signature  */
#if (JDK_VER < 110)
  arrayClazz = unhand(hclass);
#else
  arrayClazz = hclass;
#endif
  sig = cbName(arrayClazz);
  sigLength = strlen(sig);
  if ((!isSigArray(sig)) || sigLength < 2)  return NULL;
	/* Ǥʤ return */

  sig++;	/* '[' μؤ */

  /*  */
  if (sig[0] == SIGNATURE_CLASS /* 'L' */) {	/* ֥Ȥ */
    /* Lclass_name;  class_name  */
    char *findSig = strdup(sig + 1);
    char *ptr = findSig;
    while ((*ptr != SIGNATURE_ENDCLASS) && (*ptr != '\0'))  ptr++;
    *ptr = '\0';

    elemClazz = FindClassFromClass(EE(), findSig, TRUE, arrayClazz);
    if (!elemClazz)  return NULL;
  }

  if ((obj = ArrayAlloc(signature2typecode(sig), len)) == 0) {
#ifdef DEBUG
    printf("MetaClass#newArray(): ArrayAlloc() failed.\n");
    fflush(stdout);
#endif
    SignalError(0, JAVAPKG "OutOfMemoryError", 0);
    return 0;
  }
  /* η򥻥å ([Ĺ]  (ClassClass *) ) */
  if (elemClazz) {
    unhand((HArrayOfObject *)obj)->body[len] = (HObject *)elemClazz;
  }

  return (Hjava_lang_Object *)obj;
}


/*
 * 饹롣(cf. ClassLoader#defineClass())
 * ) MetaClass ˤ٤᥽åɤǤϤʤ
 */
extern struct Hjava_lang_Class *NET_shudo_moba_reflect_MetaClass_defineClass(
struct HNET_shudo_moba_reflect_MetaClass *_this,
struct Hjava_lang_ClassLoader *loader,
struct Hjava_lang_String *name,
HArrayOfByte *data, int32_t offset, int32_t length) {
  Hjava_lang_Class *clazz = NULL;

#ifdef DEBUG
  printf("MetaClass#defineClass() is called.\n");
  if (name) {
    printf(" classname: %s\n", makeCString(name));
  }
  fflush(stdout);
#endif

#if (JDK_VER < 110)
  clazz = java_lang_ClassLoader_defineClass0(loader, data, offset, length);
#else
  clazz = java_lang_ClassLoader_defineClass0(loader, name, data, offset, length);
#endif

  /* GC ǥ饹뤳Ȥɤ */
  registerGlobalHandle((Hjava_lang_Object *)clazz);

  return clazz;
}


/*
 *  writeObject() äƤ뤫֤
 */
extern /*boolean*/ int32_t NET_shudo_moba_reflect_MetaClass_hasOwnWriteObject0(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ExecEnv *ee = EE();
  ClassClass *cb = HMetaClassToClassClass(_this);
  if (find_own_rw_obj(ee, cb, 1))  return 1;	/* true */
  else  return 0;				/* false */
}


/*
 *  readObject() äƤ뤫֤
 */
extern /*boolean*/ int32_t NET_shudo_moba_reflect_MetaClass_hasOwnReadObject0(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ExecEnv *ee = EE();
  ClassClass *cb = HMetaClassToClassClass(_this);
  int32_t ret;

  if (find_own_rw_obj(ee, cb, 0))  ret = 1;	/* true */
  else  ret = 0;				/* false */
#ifdef DEBUG
  printf("%s has%s own RWObject.\n", cbName(cb), (ret?"":" *not*"));
  fflush(stdout);
#endif

  return ret;
}


/*
 * 饹ե magic number ֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_magicNumber(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  return (int32_t)JAVA_CLASSFILE_MAGIC;
}

/*
 * major version ֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_majorVersion(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  return (int32_t)cbMajorVersion(cb);
}

/*
 * minor version ֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_minorVersion(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  return (int32_t)cbMinorVersion(cb);
}

/*
 * access ե饰֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_access(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  return (int32_t)cbAccess(cb);
}

/*
 * եɤο֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_fieldCount(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);

  return cbFieldsCount(cb);
  /* return unhand(unhand(_this)->thisClass)->fields_count; */
}

/*
 * ᥽åɤο֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_methodCount(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);

  return cbMethodsCount(cb);
  /* return unhand(unhand(_this)->thisClass)->methods_count; */
}

/*
 * interface count ֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_interfaceCount(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  return (int32_t)cbImplementsCount(cb);
}


/*
 * ꤷ access modifier ĥ᥽åɤο֤
 */
extern int32_t NET_shudo_moba_reflect_MetaClass_methodCountWithAccess(struct HNET_shudo_moba_reflect_MetaClass *_this, int32_t mod) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  int methodCount = cbMethodsCount(cb);
  int ret = 0;

  struct methodblock *method;
  int i;
  int32_t invMod = ~mod;

  method = cbMethods(cb);
  for (i = cbMethodsCount(cb); --i >= 0; method++) {
    if ((method->fb.access & mod) == mod)  ret++;
  }

  return ret;
}


/*
 * classfile ΥեޥåȤ interface Ϥ롣
 */
extern void NET_shudo_moba_reflect_MetaClass_writeIFInClassfile(struct HNET_shudo_moba_reflect_MetaClass *_this, struct Hjava_io_OutputStream *out) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  int ifCount = cbImplementsCount(cb);
  short *interfaces = cbImplements(cb);
  int i;
#ifdef DEBUG
  printf("MetaClass#writeIFInClassfile() called.\n");
  fflush(stdout);
#endif

  for (i = 0; i < ifCount; i++) {
    if ((writeInt16(out, (int16_t *)(interfaces + i))) < 0)  goto exc_occurred;
  }

  return;

exc_occurred:
  THROW_IOEXCEPTION();
}

/*
 * ե֤̾
 */
extern struct Hjava_lang_String *NET_shudo_moba_reflect_MetaClass_sourceName(struct HNET_shudo_moba_reflect_MetaClass *_this) {
  ClassClass *cb = HMetaClassToClassClass(_this);
  char *sourceName = cbSourceName(cb);

  if (!sourceName)  return NULL;

  return makeJavaString(sourceName, strlen(sourceName));
}


/*
 *	Global functions
 */

/* following code is derived from
	javasrc/src/share/java/lang/ObjectInputStream.c */
struct methodblock *find_own_rw_obj(
	ExecEnv *ee, ClassClass *cb, int flag_write) {
  static unsigned long readNameAndTypeId = 0;
  static unsigned long writeNameAndTypeId = 0;

  struct methodblock *methods = cbMethods(cb);
  int methodCount = -1;
  unsigned long id = 0;
  int i;

  if(readNameAndTypeId == 0)
    readNameAndTypeId =
		NameAndTypeToHash(OWN_READOBJ, OWN_READOBJ_SIGNATURE);
  if(writeNameAndTypeId == 0)
    writeNameAndTypeId =
		NameAndTypeToHash(OWN_WRITEOBJ, OWN_WRITEOBJ_SIGNATURE);

  id = (flag_write ? writeNameAndTypeId : readNameAndTypeId);
  methodCount = cbMethodsCount(cb);
  for (i = methodCount - 1; i >= 0; i--) {
    if ((methods[i].fb.access & ACC_PRIVATE) &&		/* private */
	!(methods[i].fb.access & ACC_STATIC) &&		/* isn't static */
	(methods[i].fb.ID == id))
      return &methods[i];
  }

  return NULL;
}
