/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for marshaling */

#include <string.h>	/* for strlen() */
#include <netinet/in.h>	/* for hton[ls],ntoh[ls]() */

#include <native.h>	/* for native methods */

#include "NET_shudo_moba_io_ObjectWriter.h"
#include "NET_shudo_moba_io_ObjectReader.h"

#include <native.h>	/* for native methods */
/*#include <sys_api.h>	/* for sys*() */
#include "sysmacros_md.h"	/* for sys*() */

#include "moba.h"


#ifdef BLOCK_IO_ARRAY
#define ARRAY_SIZE_THRESHOULD	5
#define ARRAY_BLOCK_SIZE	(64 * 1024)
				/* 64KB */
#endif


/*
 * Local Functions
 */
static int writeBaseType8Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray);
static int writeBaseType16Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray);
static int writeBaseType32Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray);
static int writeBaseType64Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray);

static void readBaseType8Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray);
static void readBaseType16Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray);
static void readBaseType32Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray);
static void readBaseType64Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray);

static char *field_name(ClassClass *, int32_t);


/*
 *	Native methods for moba.ObjectOutputStream
 */

extern void NET_shudo_moba_io_ObjectWriter_writeBaseTypeArray(struct HNET_shudo_moba_io_ObjectWriter *_this, struct Hjava_lang_Object *harray, struct Hjava_lang_String *hsig) {
  ExecEnv *ee = EE();
  Hjava_io_OutputStream *out = (Hjava_io_OutputStream *)_this;
  char *sig = makeCString(hsig);
  int32_t len;

  if (!harray) {
    len = -1;
    if ((writeInt32(out, &len)) < 0)
      goto exc_occurred;
    return;
  }
  else {
    len = obj_length(harray);
    if ((writeInt32(out, &len)) < 0)
      goto exc_occurred;
  }

  if (sig[0] == SIGNATURE_ARRAY)  sig++;	/* '[' 򥹥å */

  switch (*sig) {
  case SIGNATURE_BYTE:
    if ((writeBaseType8Array(out, harray)) < 0)
      goto exc_occurred;
    break;
  case SIGNATURE_CHAR:
  case SIGNATURE_SHORT:
    if ((writeBaseType16Array(out, harray)) < 0)
      goto exc_occurred;
    break;
  case SIGNATURE_FLOAT:
  case SIGNATURE_INT:
  case SIGNATURE_BOOLEAN:
    if ((writeBaseType32Array(out, harray)) < 0)
      goto exc_occurred;
    break;
  case SIGNATURE_DOUBLE:
  case SIGNATURE_LONG:
    if ((writeBaseType64Array(out, harray)) < 0)
      goto exc_occurred;
    break;

  case SIGNATURE_ARRAY:
	/*  T_ARRAY ¸ߤʤ */
  case SIGNATURE_CLASS:
  case SIGNATURE_ENUM:
  case SIGNATURE_VOID:
  defaut:
    /* not reached */
    break;
  }

  return;

exc_occurred:
  THROW_IOEXCEPTION();
}


extern void NET_shudo_moba_io_ObjectReader_fillBaseTypeArray(struct HNET_shudo_moba_io_ObjectReader *_this, struct Hjava_lang_Object *harray, struct Hjava_lang_String *hsig) {
  ExecEnv *ee;

  Hjava_io_InputStream *in;
  int len;
  char *sig;

#ifdef DEBUG
  printf("ObjectReader#fillBaseTypeArray() is called.\n");  fflush(stdout);
#endif

  if (!harray) {
#ifdef DEBUG
    printf("object to be filled is null.\n");  fflush(stdout);
#endif
    return;
  }

  ee = EE();
  in = (Hjava_io_InputStream *)_this;
  len = obj_length(harray);
#ifdef DEBUG
  printf("length of array: %d\n", len);  fflush(stdout);
#endif

  if (len < 0)  return;	/* ǰΤ */


  sig = makeCString(hsig);
  if (sig[0] == SIGNATURE_ARRAY)  sig++;	/* '[' 򥹥å */

  switch (*sig) {
  case SIGNATURE_BYTE:
    readBaseType8Array(in, harray);
    break;
  case SIGNATURE_CHAR:
  case SIGNATURE_SHORT:
    readBaseType16Array(in, harray);
    break;
  case SIGNATURE_FLOAT:
  case SIGNATURE_INT:
  case SIGNATURE_BOOLEAN:
    readBaseType32Array(in, harray);
    break;
  case SIGNATURE_DOUBLE:
  case SIGNATURE_LONG:
    readBaseType64Array(in, harray);
    break;

  case SIGNATURE_ARRAY:
	/*  T_ARRAY ¸ߤʤ */
  case SIGNATURE_CLASS:
  case SIGNATURE_ENUM:
  case SIGNATURE_VOID:
  defaut:
    /* not reached */
    break;
  }


  /* 㳰 throw ľ */
  if (exceptionOccurred(ee)) {
    exceptionThrow(ee, ee->exception.exc);
  }
#ifdef DEBUG
  printf("fillBaseTypeArray is done.\n");
  fflush(stdout);
#endif
}


/*
 * Local Functions
 */

static int writeBaseType8Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray) {
  return writeArrayOfByte(out, (HArrayOfByte *)harray, 0, obj_length(harray));
  /* returns : 0, : -1 */
}

static int writeBaseType16Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray) {
  int len, i;
  int16_t *body;

  if (harray) {
    len = obj_length(harray);
    body = unhand((HArrayOfShort *)harray)->body;

#ifdef BLOCK_IO_ARRAY
    if (len >= ARRAY_SIZE_THRESHOULD) {
      int bound;
#ifndef BIG_ENDIAN
      int16_t *buf;
      if (!(buf = (int16_t *)sysMalloc(ARRAY_BLOCK_SIZE)))
	goto normal_io;
#endif

      while (len > 0) {
	bound = min(len, ARRAY_BLOCK_SIZE / 2);
#ifdef BIG_ENDIAN
	if ((writeBytes(out, (char *)body, 2 * bound)) < 0)
	  return -1;
#else
	for (i = 0; i < bound; i++)
	  buf[i] = (int16_t)htons((unsigned short)body[i]);
	if ((writeBytes(out, (char *)buf, 2 * bound)) < 0)
	  return -1;
#endif
	len -= bound;
      }

#ifndef BIG_ENDIAN
      sysFree(buf);
#endif
      return 0;
    }
#endif

  normal_io:
    for (i = 0; i < len; i++) {
      if ((writeInt16(out, body + i)) < 0)
	return -1;
    }
  }
  return 0;
}

static int writeBaseType32Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray) {
  int len, i;
  int32_t *body;

  if (harray) {
    len = obj_length(harray);
    body = unhand((HArrayOfInt *)harray)->body;

#ifdef BLOCK_IO_ARRAY
    if (len >= ARRAY_SIZE_THRESHOULD) {
      int bound;
#ifndef BIG_ENDIAN
      int32_t *buf;
      if (!(buf = (int32_t *)sysMalloc(ARRAY_BLOCK_SIZE)))
	goto normal_io;
#endif

      while (len > 0) {
	bound = min(len, ARRAY_BLOCK_SIZE / 4);
#ifdef BIG_ENDIAN
	if ((writeBytes(out, (char *)body, 4 * bound)) < 0)
	  return -1;
#else
	for (i = 0; i < bound; i++)
	  buf[i] = (int32_t)htonl((unsigned long)body[i]);
	if ((writeBytes(out, (char *)buf, 4 * bound)) < 0)
	  return -1;
#endif
	len -= bound;
      }

#ifndef BIG_ENDIAN
      sysFree(buf);
#endif
      return 0;
    }
#endif

  normal_io:
    for (i = 0; i < len; i++) {
      if ((writeInt32(out, body + i)) < 0)
	return -1;
    }
  }
  return 0;
}

static int writeBaseType64Array(struct Hjava_io_OutputStream *out, struct Hjava_lang_Object *harray) {
  int len, i;
  int64_t *body;

  if (harray) {
    len = obj_length(harray);
    body = unhand((HArrayOfLong *)harray)->body;

#ifdef BLOCK_IO_ARRAY
    if (len >= ARRAY_SIZE_THRESHOULD) {
      int bound;
#ifndef BIG_ENDIAN
      int64_t *buf;
      if (!(buf = (int64_t *)sysMalloc(ARRAY_BLOCK_SIZE)))
	goto normal_io;
#endif

      while (len > 0) {
	bound = min(len, ARRAY_BLOCK_SIZE / 8);
#ifdef BIG_ENDIAN
	if ((writeBytes(out, (char *)body, 8 * bound)) < 0)
	  return -1;
#else
	for (i = 0; i < bound; i++)
	  buf[i] = hton64((uint64_t)body[i]);
	if ((writeBytes(out, (char *)buf, 8 * bound)) < 0)
	  return -1;
#endif
	len -= bound;
      }

#ifndef BIG_ENDIAN
      sysFree(buf);
#endif
      return 0;
    }
#endif

  normal_io:
    for (i = 0; i < len; i++) {
      if ((writeInt64(out, body + i)) < 0)
	return -1;
    }
  }
  return 0;
}


static void readBaseType8Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray) {
  readArrayOfByte(in, (HArrayOfByte *)harray, 0, obj_length(harray));
}

static void readBaseType16Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray) {
  int16_t *body, data;
  int len, i;

  if (harray) {
    len = obj_length(harray);
    body = unhand((HArrayOfShort *)harray)->body;

    readBytes(in, (char *)body, 2 * len);
    /* byte order Ѵ */
    for (i = 0; i < len; i++) {
      body[i] = (int16_t)ntohs((unsigned short)body[i]);
    }
  }
}

static void readBaseType32Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray) {
  int32_t *body, data;
  int len, i;

  if (harray) {
    len = obj_length(harray);
    body = unhand((HArrayOfInt *)harray)->body;

    readBytes(in, (char *)body, 4 * len);
    /* byte order Ѵ */
    for (i = 0; i < len; i++) {
      body[i] = (int32_t)ntohl((unsigned long)body[i]);
    }
  }
}

static void readBaseType64Array(struct Hjava_io_InputStream *in, struct Hjava_lang_Object *harray) {
  int64_t *body, data;
  int len, i;

  if (harray) {
    len = obj_length(harray);
    body = unhand((HArrayOfLong *)harray)->body;

    readBytes(in, (char *)body, 8 * len);
    /* byte order Ѵ */
    for (i = 0; i < len; i++) {
      body[i] = (int64_t)ntoh64((uint64_t)body[i]);
    }
  }
}



/* following code is derived from
	javasrc/src/share/java/lang/ObjectInputStream.c */
extern /*boolean*/ int32_t NET_shudo_moba_io_ObjectWriter_invokeOwnObjectWriter(struct HNET_shudo_moba_io_ObjectWriter *_this, struct Hjava_lang_Class *classclass,struct Hjava_lang_Object *obj) {
  ClassClass *cb = NULL;
  ExecEnv *ee = EE();
  struct methodblock* method;

#if (JDK_VER < 110)
  cb = unhand(classclass);
#else
  cb = classclass;
#endif

  method = find_own_rw_obj(ee, cb, 1);
  if (!method || exceptionOccurred(ee))  return 0;

  if (java_monitor) {
    struct methodblock *caller;
    long start_time, end_time;
    start_time = now();
    do_execute_java_method(ee, obj,
		OWN_WRITEOBJ, OWN_WRITEOBJ_SIGNATURE, method, 0,
		/* argument */ _this, NULL);
    end_time = now();
    caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
    if (!caller)  caller = ee->current_frame->current_method;
    java_mon(caller, method, end_time-start_time);
  }
  else {
    do_execute_java_method(ee, obj,
		OWN_WRITEOBJ, OWN_WRITEOBJ_SIGNATURE, method, 0,
		/* argument */ _this, NULL);
  }
  return 1;
}


/* following code is derived from
	javasrc/src/share/java/lang/ObjectInputStream.c */
extern /*boolean*/ int32_t NET_shudo_moba_io_ObjectReader_invokeOwnObjectReader(struct HNET_shudo_moba_io_ObjectReader *_this, struct Hjava_lang_Class *classclass, struct Hjava_lang_Object *obj) {
  ClassClass *cb = NULL;
  ExecEnv *ee = EE();
  struct methodblock* method;

#if (JDK_VER < 110)
  cb = unhand(classclass);
#else
  cb = classclass;
#endif

  method = find_own_rw_obj(ee, cb, 0);
  if (!method || exceptionOccurred(ee))  return 0;

  if (java_monitor) {
    struct methodblock *caller;
    long start_time, end_time;
    start_time = now();
    do_execute_java_method(ee, obj,
		OWN_READOBJ, OWN_READOBJ_SIGNATURE, method, 0,
		/* argument */ _this, NULL);
    end_time = now();
    caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
    if (!caller)  caller = ee->current_frame->current_method;
    java_mon(caller, method, end_time-start_time);
  }
  else {
    do_execute_java_method(ee, obj,
		OWN_READOBJ, OWN_READOBJ_SIGNATURE, method, 0,
		/* argument */ _this, NULL);
  }
  return 1;
}
