/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for thread streaming */

#include <stdio.h>	/* for NULL */
#include <string.h>	/* for memcpy() */
#include <sys/types.h>	/* for open(),... */
#include <sys/stat.h>	/* for open(),... */
#include <fcntl.h>	/* for open(),... */
#include <unistd.h>	/* for sleep() */

#include <native.h>	/* for native methods */
#include <sys_api.h>	/* for type sys_thread_t, sys*() */

#if 0
#include "threads.h"	/* for JavaStackSize, THR_SYSTEM, THR_USER */
#else
#include "java_lang_Thread.h"	/* for struct Classjava_lang_Thread */
#include "java_lang_ThreadGroup.h"
#define THR_SYSTEM	0
#define THR_USER	1
#define threadSetPriority(hThread, i) \
	sysThreadSetPriority((sys_thread_t *)(unhand(hThread)->PrivateInfo), i)
#define threadInit(hThread, sb) \
	sysThreadInit((sys_thread_t *)(unhand(hThread)->PrivateInfo), sb)
#define threadExit()	sysThreadExit()
extern long JavaStackSize;
extern long ProcStackSize;
#endif

#include "java_lang_Object.h"
#include "NET_shudo_moba_lang_MobaThread.h"

#include "moba.h"

#include "stack.h"


/*
 * Variables
 */
#define EXCEPTION_BUF_SIZE	100
static char exception_msg_buf[EXCEPTION_BUF_SIZE];


/*
 * Local Functions
 */
static void MobaThreadRT(Hjava_lang_Thread *p);
static unsigned char *returnPC(JavaFrame *frame);

static void dumpAllFrames(FILE *out, JavaFrame *topFrame);
static void dumpFrame(FILE *out, JavaFrame *frame);


/*
 * åɤ˴ؤ serialize
 */
extern void NET_shudo_moba_lang_MobaThread_writeContextTo(struct HNET_shudo_moba_lang_MobaThread *_this, struct HNET_shudo_moba_io_ObjectWriter *objOut) {
/*
  ClassClass *mobaThreadCb = handleToClassClass(_this);
  ClassClass *threadCb = cbSuperclass(mobaThreadCb);
*/
  ExecEnv *ee = (ExecEnv *)(unhand(_this)->eetop);
  JavaFrame *frame;
  struct methodblock *method;

  uint16_t nFrames;
  uint32_t lastpc_offset;

  Hjava_io_OutputStream *out = (Hjava_io_OutputStream *)objOut;


#ifdef DEBUG
  printf("writeContextTo() called. (obj: 0x%x)\n", (int)_this);
#endif

  /* JavaFrame 򿷤˥åѤ */
  stack_init();
  nFrames = 0;
  frame = ee->current_frame;
#if 0
  /* Υɤͭˤ examples/ThreadWriterTest.java ưʤ
	Oct 6 1997 */
  dumpAllFrames(stdout, frame);
  fflush(stdout);
#endif
  while (frame) {	/* assume initial_frame->prev == NULL */
    if (frame->current_method) {
      nFrames++;
      stack_push((void *)frame);
    }
    else {
      /* current_method ʤ frame
	 Ĥޤ main ᥽åɤ̵ */
      break;
    }
    frame = frame->prev;
  }

#ifdef DEBUG
  printf("# of frames: %d\n", (int)nFrames);
  fflush(stdout);
#endif
  if ((writeInt16(out, (int16_t *)&nFrames)) < 0)
    goto exc_occurred;


  /* Ť JavaFrame é */
  while (!stack_is_empty()) {
    int frameNo = nFrames - stack_depth();	/* for debug */

    frame = (JavaFrame *)stack_pop();
    method = frame->current_method;
#ifdef DEBUG
    printf("\n");
    printf("writing JavaFrame (%d)\n", frameNo);
    fflush(stdout);
#endif

    /* JavaFrame (Ϣ) 񤭽Ф */
#ifdef EASY_READING
    {
      uint16_t frameNo16 = (uint16_t)frameNo;
      if ((writeChar(out, '{')) < 0)  goto exc_occurred;
      if ((writeInt16(out, (int16_t *)&frameNo16)) < 0)  goto exc_occurred;
      if ((writeChar(out, '}')) < 0)  goto exc_occurred;
    }
#endif

    /* 饹᥽å */
    {
      uint16_t methodAcc = method->fb.access;

      if ((writeClassAndMethod(out, method)) < 0)  goto exc_occurred;
      if ((writeInt16(out, (int16_t *)&methodAcc)) < 0)  goto exc_occurred;
	/* access flag */
    }
    /* Ѱ or Τޤޤͤǽ񤭽Ф */
    /* returnpc, optop, lastpc, mon_starttime */
    {
      uint32_t returnpc_offset;		/* 4 bytes */
      int16_t optop_offset;		/* 2 bytes ǽʬ (< 2000) */
      /* uint32_t lastpc_offset;		/* 4 bytes */

      returnpc_offset = pc2offset(method, frame->returnpc);
      optop_offset = sp2offset(frame, frame->optop);
      lastpc_offset = pc2offset(method, frame->lastpc);
#ifdef DEBUG
      printf("returnpc: %d\n", returnpc_offset);
      printf("optop:    %d\n", (int)optop_offset);
      printf("lastpc:   %d\n", lastpc_offset);
      fflush(stdout);
#endif

      if ((writeInt32(out, (int32_t *)&returnpc_offset)) < 0)
	goto exc_occurred;
      if ((writeInt16(out, &optop_offset)) < 0)  goto exc_occurred;
      if ((writeInt32(out, (int32_t *)&lastpc_offset)) < 0)  goto exc_occurred;

#if 0
      /* JavaFrame.mon_starttime */
      if ((writeInt32(out, (int32_t *)&frame->mon_starttime)) < 0)
	goto exc_occurred;
#endif
    }
    /* constant_pool */
    /*  returnpc ˰¸
     *  returnpc is 0: (struct methodblock *)
     *       is not 0: (cp_item *)
     *  ExecuteJava() (in javasrc/share/java/runtime/executeJava.c) 
     */
    {
#if (JDK_VER < 110)
      if ((writeChar(out, CP_CP)) < 0)  goto exc_occurred;
#else
      if (frame->returnpc == 0) {
	/* (struct methodblock *)frame->constant_pool */
	struct methodblock *targetMethod =
		(struct methodblock *)frame->constant_pool;
	if ((writeChar(out, CP_METHOD)) < 0)  goto exc_occurred;

	if ((writeClassAndMethod(out, targetMethod)) < 0)  goto exc_occurred;
      }
      else {
	if ((writeChar(out, CP_CP)) < 0)  goto exc_occurred;
      }
#endif /* JDK_VER < 110 */
    }
    /* local ѿ */
#ifdef DEBUG
    printf("writing local vars.\n");
    printf(" (#localvar: %d, nlocals: %d)\n",
	LocalVarTblLen(method), method->nlocals);
    fflush(stdout);
#endif
    {
      long slot;  uint16_t slot16;
      int i;

      int lvTblLen = LocalVarTblLen(method);
      if (lvTblLen > 0) {	/* LocalVariableTable ĥ᥽å */
	char *sig;
  #ifdef DEBUG
	char *name;
	long pc0, length;
  #endif

	for (i = 0; i < LocalVarTblLen(method); i++) {
	  slot = LocalVarSlot(method, i);
	  sig = LocalVarSig(method, i);
  #ifdef DEBUG
	  name = LocalVarName(method, i);
	  pc0 = LocalVarStartPC(method, i);
	  length = LocalVarLengthPC(method, i);
	  printf(" [%d] %s %s\tslot:%ld\tPC:%ld(%ld)\n",
		i, sig, name, slot, pc0, length);
	  fflush(stdout);
  #endif
	  if (isLocalVarAlive(method, i, lastpc_offset)) {
	    /* local var. is alive */
  #ifdef DEBUG
	    printf("    is alive.\n");  fflush(stdout);
  #endif
	    /* slot */
	    slot16 = (uint16_t)slot;
  #ifdef EASY_READING
	    if ((writeChar(out, '[')) < 0)  goto exc_occurred;
  #endif
	    if ((writeInt16(out, (int16_t *)&slot16)) < 0)  goto exc_occurred;
  #ifdef EASY_READING
	    if ((writeChar(out, ']')) < 0)  goto exc_occurred;
  #endif
	    /* content */
	    if ((writeStackItem(objOut, frame->vars + slot, sig)) < 0)
	      goto exc_occurred;
	  }	/* local var. is alive. */
	}
      }
      else {	/* LocalVariableTable ʤ᥽å */
	for (slot = 0; slot < method->nlocals; slot++) {
  #ifdef DEBUG
	  printf(" slot:%ld\n", slot);
	  fflush(stdout);
  #endif
	  /* slot */
	  slot16 = (uint16_t)slot;
  #ifdef EASY_READING
	  if ((writeChar(out, '[')) < 0)  goto exc_occurred;
  #endif
	  if ((writeInt16(out, (int16_t *)&slot16)) < 0)  goto exc_occurred;
  #ifdef EASY_READING
	  if ((writeChar(out, ']')) < 0)  goto exc_occurred;
  #endif
	  /* content */
	  if ((writeStackItemAuto(objOut, frame->vars + slot)) < 0)
	    goto exc_occurred;
	}
      }		/* if (lvTblLen > 0) */


      {
	/* terminate local vars. */
	uint16_t term = 0xffff;
#ifdef EASY_READING
	if ((writeChar(out, '[')) < 0)  goto exc_occurred;
#endif
	if ((writeInt16(out, (int16_t *)&term)) < 0)  goto exc_occurred;
#ifdef EASY_READING
	if ((writeChar(out, ']')) < 0)  goto exc_occurred;
#endif
      }
    }
    /* å, JavaFrame.optop */
    {
      stack_item *item = frame->ostack;
#ifdef DEBUG
      int i = 0;
#endif

#ifdef DEBUG
      if (item < frame->optop) {
	printf("writing stack items.\n");
	fflush(stdout);
      }
#endif
      while (item < frame->optop) {
#ifdef DEBUG
	printf("writing frame->ostack[%d]\n", i++);
	fflush(stdout);
#endif
	if ((writeStackItemAuto(objOut, item)) < 0)  goto exc_occurred;
	item++;
      }
    }
  }	/* JavaFrame é */

#ifdef DEBUG
  printf("writeContextTo() done.\n");
  fflush(stdout);
#endif

  return;

exc_occurred:
  THROW_IOEXCEPTION();
}


/*
 * åɤ˴ؤ de-serialize
 */
extern void NET_shudo_moba_lang_MobaThread_readContextFrom(struct HNET_shudo_moba_lang_MobaThread *_this, struct HNET_shudo_moba_io_ObjectReader *objIn, struct Hjava_lang_Thread *hTarget) {
/*
  ClassClass *mobaThreadCb = handleToClassClass(_this);
  ClassClass *threadCb = cbSuperclass(mobaThreadCb);
*/
  ExecEnv *ee = EE();
  ExecEnv *targetEE;
  JavaFrame *frame;
#if 0
  JavaFrame *initialNullFrame;
#endif
  struct methodblock *method;
  ClassClass *clazz;

  uint16_t nFrames;
  int frameCount;

  Hjava_io_InputStream *in = (Hjava_io_InputStream *)objIn;


#ifdef DEBUG
  printf("readContextFrom() is called.\n");
  printf("hTarget: 0x%x\n", hTarget);
  {
    HArrayOfChar *hName = unhand(hTarget)->name;
    unicode *name = unhand(hName)->body;
    int32_t priority = unhand(hTarget)->priority;

    printf("hTarget...name: ");
    outUnicode(stdout, name);
    printf("\n");
    printf("htarget...priority: %d\n", (int)priority);
  }

  fflush(stdout);
#endif
  targetEE = (ExecEnv *)sysMalloc(sizeof(ExecEnv));

  InitializeExecEnv(targetEE, (JHandle *)NULL);
  /*
   * InitializeExecEnv(targetEE, (JHandle *)hTarget ȤϤǤʤ
   * InitializeExecEnv()  unhand(ä)->initiali_stack_memory 
   * Ƥ뤳ȤԤΤǡ
   */
  unhand(hTarget)->eetop = (int32_t)targetEE;
  targetEE->thread = (JHandle *)hTarget;

#if 0
  /* fake for JDK profiling */
  /* allocate null frame */
  initialNullFrame = allocNewJavaFrame(NULL, targetEE);
#endif


  nFrames = (uint16_t)readInt16(in);
#ifdef DEBUG
  printf("# of frames: %d\n", (int)nFrames);
  fflush(stdout);
#endif


  for (frameCount = 0; frameCount < nFrames; frameCount++) {
#ifdef DEBUG
    printf("\n");
    printf("reading stack frame(%d):\n", frameCount);
    fflush(stdout);
#endif
    /* JavaFrame (Ϣ) ƹ */
#ifdef EASY_READING
    {
      uint16_t frameNo16 = -1;
      readChar(in);
      frameNo16 = (uint16_t)readInt16(in);
      readChar(in);
    }
#endif

    /* 饹᥽å */
    {
      uint16_t methodAcc;
      int i;

      method = readClassAndMethod(objIn);
      methodAcc = (uint16_t)readInt16(in);	/* access flag */

      if (!method)  return;


      /* access flag Ĵ٤ */
#define ACC_EQ_MASK	(ACC_NATIVE | ACC_MACHINE_COMPILED)
      if ((methodAcc & ACC_EQ_MASK) != (method->fb.access & ACC_EQ_MASK)) {
	jio_snprintf(exception_msg_buf, EXCEPTION_BUF_SIZE,
		"%s#%s", cbName(method->fb.clazz), method->fb.name);
	SignalError(0, MOBA_PACKAGE "FieldNotCompatibleException",
		exception_msg_buf);
	return;
      }
    } /* 饹᥽å */


    /* JavaFrame 򥻥åȥå */
    frame = allocNewJavaFrame(method, targetEE);

    /* Ѱ or Τޤޤͤǽ񤭽Ф */
    /* returnpc, optop, lastpc, mon_starttime */
    /* constant_pool  allocNewJavaFrame() ꤵƤ */
    {
      uint32_t returnpc_offset;		/* 4 bytes */
      int16_t optop_offset;		/* 2 bytes */
      uint32_t lastpc_offset;		/* 4 bytes */

      returnpc_offset = (uint32_t)readInt32(in);
      optop_offset = readInt16(in);
      lastpc_offset = (uint32_t)readInt32(in);
#ifdef DEBUG
      printf("returnpc: %d\n", returnpc_offset);
      printf("optop:    %d\n", (int)optop_offset);
      printf("lastpc:   %d\n", lastpc_offset);
      fflush(stdout);
#endif

      frame->returnpc = offset2pc(method, returnpc_offset);
      frame->optop = offset2sp(frame, optop_offset);
      frame->lastpc = offset2pc(method, lastpc_offset);

#if 0
      /* JavaFrame.mon_starttime */
      frame->mon_starttime = (int)readInt32(in);
#endif
    }
    /* constant_pool */
    {
      struct methodblock *method;
      char cpType = readChar(in);

#if (JDK_VER >= 110)
      switch(cpType) {
      case CP_CP:
	/* ɬפʤ  allocNewJavaFrame() ꤵƤΤ */
	break;
      case CP_METHOD:
	method = readClassAndMethod(objIn);
	frame->constant_pool = (cp_item_type *)method;
#ifdef DEBUG
	printf("constant_pool is pointing method: 0x%x (%s#%s)\n",
		(int)method, cbName(method->fb.clazz), method->fb.name);
#endif
	break;
      default:
	SignalError(0, MOBA_PACKAGE "StreamFormatException",
			"cp type is expected.");
	return;
	break;
      }
#endif /* JDK_VER >= 110 */
    }
    /* local ѿ */
    {
      uint16_t slot = 0;
      int itemType;

#ifdef DEBUG
      printf("reading local vars.\n");
      fflush(stdout);
#endif
      while (1) {
#ifdef EASY_READING
	readChar(in);
#endif
	slot = (uint16_t)readInt16(in);
#ifdef EASY_READING
	readChar(in);
#endif
	if (slot == (uint16_t)0xffff) {
	  break;
	}
#ifdef DEBUG
	printf("slot:%d\n", (int)slot);
	fflush(stdout);
#endif

	itemType = readStackItem(objIn, &(frame->vars[slot]));
#ifdef DEBUG
	{
	  stack_item item = frame->vars[slot];

	  printf("\tlocal var[%d]: ", (int)slot);
	  switch (itemType) {
	  case STACK_ITEM_HANDLE:
	    if (item.h) {
	      if (obj_flags(item.h) == T_NORMAL_OBJECT)
		printf("%s", cbName(obj_classblock(item.h)));
	      else
		printf("array");
	    }
	    else
	      printf("(null)");
	    break;
	  case STACK_ITEM_IMM32:
	    printf("%d", item.i);
	    break;
	  case STACK_ITEM_IMM64:
	    printf("0x%x, 0x%x", frame->vars[slot].i, frame->vars[slot+1].i);
	    break;
	  default:
	    printf("0x%x", item.i);
	    break;
	  }
	  printf("\n");
	  fflush(stdout);
	}
#endif
      }
    }
    /* å */
    {
      stack_item *stack = frame->ostack;

      while (stack < frame->optop) {
	readStackItem(objIn, stack);
	stack++;
      }
    }
    /* ¸פ JavaFrame Ф */
    {
      /* prev */
      /* allocNewJavaFrame() ꤵ */

      /* vars */
      frame->vars = ((stack_item *)frame) - (method->nlocals);

      /* javastack */
      /* frame ΥݻꤵƤ */
    }
  }	/* JavaFrame é */


#ifdef DEBUG
  printf("\nExecEnv->current_frame:\n");
  dumpAllFrames(stdout, targetEE->current_frame);
  fflush(stdout);
#endif

#ifdef DEBUG
  printf("readContextFrom() done.\n");
  fflush(stdout);
#endif
}


/**
 * ߤ PC ˴ؤ롣
 */
extern struct Hjava_lang_String *NET_shudo_moba_lang_MobaThread_currentPCString(struct HNET_shudo_moba_lang_MobaThread *_this) {
  ExecEnv *ee = (ExecEnv *)((unhand(_this))->eetop);
  JavaFrame *frame = ee->current_frame;
  struct methodblock *method = frame->current_method;
  unsigned char *pc = frame->lastpc;

#define PC_BUF_SIZE 100
  static char buf[PC_BUF_SIZE];


  pc2string(pc, method, buf, buf + PC_BUF_SIZE);

  return makeJavaString(buf, strlen(buf));
}


/*
 * internalize 줿 MobaThread ư롣
 */
extern void NET_shudo_moba_lang_MobaThread_boot(struct HNET_shudo_moba_lang_MobaThread *_this) {
  struct Classjava_lang_Thread *tid;

  monitorEnter(obj_monitor(_this));

  tid = (struct Classjava_lang_Thread *)unhand(_this);

  if (tid->PrivateInfo) {
    SignalError(0, "java/lang/IllegalThreadStateException", 0);
  }
  else {
    int flag = (tid->daemon ? THR_SYSTEM : THR_USER);
    if (threadCreate(_this, flag, ProcStackSize,
		(void *(*)(void *))MobaThreadRT)) {
      SignalError(0, "java/lang/OutOfMemoryError", 0);
    }
    else {
      threadSetPriority(_this, tid->priority);
      /* threadResume(thread); */
    }
  }

  monitorExit(obj_monitor(_this));

#ifdef DEBUG
  {
    HArrayOfChar *name = unhand(_this)->name;
    unicode *namebody = unhand(name)->body;
    printf("MobaThread[");
    outUnicode(stdout, namebody);
    printf("] is booted.\n");
    fflush(stdout);
  }
#endif
}


/*
 * Local Functions
 */

/*
 * ThreadRT0() (in javasrc/src/share/java/lang/thread.c)
 *  MobaThread 
 */
static void MobaThreadRT(Hjava_lang_Thread *thread) {
  ExecEnv ee;
  JavaFrame *frame;
  struct methodblock *method;
  unsigned char *pc = NULL;

  /* 㳰 */
  int exc_occurred = 0;
	/*  exceptionOccurred(&ee) η̤Ʊݤ */
  JHandle *exc = NULL;


  ExecEnv *givenEE = (ExecEnv *)unhand(thread)->eetop;
	/* ThreadRT0() Ȥϰۤʤ Ǥ˺äƤΤѤ */
  if (givenEE) {
    /* Thread#eetop 򤳤δؿ (MobaThreadRT())  auto ѿ ee ˤ */
    memcpy(&ee, givenEE, sizeof(ExecEnv));
    unhand(thread)->eetop = (int32_t)&ee;

    sysFree(givenEE);
	/*  free(): readContextFrom()  malloc() Ƥ */
  }
  else {
    InitializeExecEnv(&ee, (JHandle *)thread);
  }


  frame = ee.current_frame;
  method = frame->current_method;


  threadInit(thread, (stackp_t) &thread);
  if (unhand(thread)->stillborn)
    threadExit();

  exceptionClear(&ee);


  /* Thread#suspend0  return Ƽ¹ԺƳ */
  pc = returnPC(frame);


#ifdef DEBUG
  {
#define RT_PC_BUFSIZE	50
    char buf[RT_PC_BUFSIZE];
    pc2string(pc, method, buf, buf + RT_PC_BUFSIZE);
    printf("pc: ");
    outPC(stdout, pc, method);
    printf("\n");
    printf("MobaThreadRT(): %s\n", buf);

    printf("\n");
    fflush(stdout);
  }
#endif

  while (1) {
    bool_t execRet;

#if 0
    {
      /* ExecuteJava()  threadCheckStack() 
	 false ֤ƤΤǤϤʤȵä */
      int ret;
      ret = sysThreadCheckStack();
      printf("threadCheckStack() returns: %s\n", (ret ? "true":"false"));
      fflush(stdout);
    }
#endif

#ifdef DEBUG
    printf("call ExecuteJava(");
    outPC(stdout, pc, frame->current_method);
    printf(")\n");
    fflush(stdout);
#endif
    {
      /* fake for JDK profiling */
      int isFrameInitial = !frame->prev;
      JavaFrame tempFrame;

      if (isFrameInitial) {
	memset(&tempFrame, 0, sizeof(JavaFrame));
	frame->prev = &tempFrame;
      }
      /* fake end */

      /* for profiling */
      if (java_monitor)  frame->mon_starttime = now();

      execRet = ExecuteJava(pc, &ee);
	/*  return Ƥߤʤ... */
	/* ExecuteJava()  initial frame ̤ return Ƥޤ */

      /* fake for JDK profiling */
      if (isFrameInitial)  frame->prev = NULL;
      /* fake end */
    }
#ifdef DEBUG
    printf("ExecuteJava() returns: %s\n", (execRet ? "TRUE":"FALSE"));
    fflush(stdout);
#endif

    /*exc_occurred = exceptionOccurred(&ee);*/
    exc_occurred = !execRet;	/* This way takes more low cost */
    exc = ee.exception.exc;

    if (exc_occurred) {	/* 㳰ȯ */
      unsigned char *new_pc;

#ifdef DEBUG
      printf("in MobaThreadRT(), exception(kind: %d) occurred:\n",
		ee.exceptionKind);
      fflush(stdout);

      if (ee.exceptionKind == EXCKIND_THROW) {
	JHandle *exc = ee.exception.exc;
	printf("  Exception: ");
	if (exc)
	  printf("%s\n", cbName(obj_classblock(exc)));
	else
	  printf("(null)\n");
      }
      exceptionDescribe(&ee);
      printf("\n");
      fflush(stderr);  fflush(stdout);
#endif

    handle_exception:
      pc = frame->lastpc;
      new_pc = ProcedureFindThrowTag(&ee, frame, exc, pc);
      if (new_pc) {	/* exception handler is found */
#ifdef DEBUG
	printf("Exception handler is found (PC: 0x%x).\n", (int)new_pc);
	fflush(stdout);
#endif
	exceptionClear(&ee);  exc_occurred = 0;
	frame->optop = frame->ostack;	/* reset to empty */

	/* push exception object */
	frame->optop->h = exc;
	frame->optop++;

	pc = new_pc;

	continue;
      }
      else {		/* not found */
#ifdef DEBUG
	printf("Exception handler is *not* found.\n");
	fflush(stdout);
#endif
	/* 㳰ϥɥõƤȤΤ
	   ʲνǹԤɬפ */
	if (frame->monitor)
	  monitorExit(obj_monitor(frame->monitor));
	if (java_monitor) {
	  struct methodblock *mb = frame->current_method;
	  if (mb && frame->prev) {
		/* assume initial_frame->prev == NULL */
	    int time = frame->mon_starttime;
	    time = ((time != 0) ? (now() - time) : 0);
		// ᥽å or ExecuteJava() θƤӽФ
		// frame->mon_starttime Ƥʤθ
#ifdef DEBUG
	    {
	      struct methodblock *caller = frame->prev->current_method;
	      struct methodblock *callee = mb;
	      printf("call java_mon(caller:%s, callee:%s, time:%d)\n",
		(caller ? (caller->fb.name):"(null)"),
		(mb ? (mb->fb.name):"(null)"), time);
	      fflush(stdout);
	    }
#endif
	    java_mon(frame->prev->current_method, mb, time);
	  }
	}

	/* fall through */
      }
    }	/* if (exc_occurred) */


    /* ˾ޤ opc_return Ф */
    frame = ee.current_frame = ee.current_frame->prev;

    /* execution finishes */
    if (!frame)  break;
	/* assume initial_frame->prev == NULL */
    if (!(pc = returnPC(frame)))  break;
#ifdef DEBUG
    printf("MobaThreadRT(): frame: 0x%x -> 0x%x.\n",
		(int)frame, (int)frame->prev);
    fflush(stdout);
#endif
    if (exc_occurred)	/* exception ̤ */
      goto handle_exception;
#ifdef DEBUG
    printf("now returnd to previous frame.\n\n");
    fflush(stdout);
#endif
  }	/* while (1) */


#ifdef DEBUG
  printf("MobaThreadRT(): Execution done.\n");
  fflush(stdout);
#endif

  if (exc_occurred) {	/* ̤㳰 */
#ifdef DEBUG
    printf("An exception is uncauted.\n");
    /* exceptionDescribe(&ee); */
    printf("  parent: 0x%x\n", (int)(unhand(thread)->group));
    fflush(stdout);
#endif
    if (unhand(thread)->group) {
      exceptionClear(&ee);
      execute_java_dynamic_method(&ee, (void *)(unhand(thread)->group),
	"uncaughtException", "(Ljava/lang/Thread;Ljava/lang/Throwable;)V",
	thread, exc);
    }
  }

  unhand(thread)->stillborn = 1;	/* true */
  /* done. */

#ifdef DEBUG
  printf("Now call Thread#exit()\n");  fflush(stdout);
#endif
  {
    int count = 3;
    while ((unhand(thread)->group) && (count-- > 0)) {
      exceptionClear(&ee);
    execute_java_dynamic_method(&ee, (void *)thread, "exit", "()V");
    }
    sysAssert(unhand(thread)->group == NULL);
  }

#ifdef DEBUG
  printf("Now call threadExit()\n");  fflush(stdout);
#endif
  threadExit();
}


/*
 * opcode "return" ǽ롣
 * Ƥֻǡframe  return ΤΤˤƤ
 *
 * ExecuteJava() in javasrc/src/share/java/runtime/executeJava.c 
 * auto ѿ initial_frame 뤻ǡopcode "return " 
 * ExecuteJava()  return ƤޤʤΤ return ǽ
 *
 * returns: PC
 */
static unsigned char *returnPC(JavaFrame *frame) {
  if (frame->returnpc == 0) {
    struct methodblock *method = frame->current_method;
    struct methodblock *targetMethod =
	(struct methodblock *)frame->constant_pool;
    unsigned char *lastpc = frame->lastpc;

    if (!method) {	/* frame->current_method == NULL к */
      if (lastpc)
	method = pc2method(lastpc);
      else
	return NULL;
    }

    frame->returnpc = lastpc + 3;
	/* frame->returnpc = nextPC(lastpc); */
	/* ǤܡɤκŬˤäԹ礬롣 */
    frame->constant_pool = cbConstantPool(fieldclass(&method->fb));
#ifdef DEBUG
    if (method && lastpc && frame->returnpc) {
      uint32_t off0, off1;
      off0 = pc2offset(method, lastpc);
      off1 = pc2offset(method, frame->returnpc);
      printf("pc(offset): %u -> %u  (in returnPC())\n", off0, off1);
      fflush(stdout);
    }
#endif

#if 0	/* comment func. call FixupQuickInvocation() out. */
#if (JDK_VER >= 110)
#ifdef DEBUG
    printf("call FixupquickInvoction():\n");
    printf(" lastpc: ");  outPC(stdout, lastpc, method);  printf("\n");
    printf(" current_method: 0x%x ", (int)method);
    outMethod(stdout, method);
    printf("\n");
    printf(" target_method:  0x%x ", (int)targetMethod);
    outMethod(stdout, targetMethod);
    printf("\n");
    fflush(stdout);
#endif
#ifdef DEBUG
    printf("before FixupQuickInvocation(): *lastpc: %3d %3d %3d %3d %3d\n",
	(int)*lastpc, (int)*(lastpc+1), (int)*(lastpc+2),
	(int)*(lastpc+3), (int)*(lastpc+4));
#endif
    FixupQuickInvocation(lastpc, method, targetMethod);
#ifdef DEBUG
    printf("after  FixupQuickInvocation(): *lastpc: %3d %3d %3d %3d %3d\n",
	(int)*lastpc, (int)*(lastpc+1), (int)*(lastpc+2),
	(int)*(lastpc+3), (int)*(lastpc+4));
#endif
#endif	/* JDK_VER >= 110 */
#endif	/* comment func. call FixupQuickInvocation() out. */
  }

#if 0
  printf("returnPC() returns: 0x%x\n", (int)frame->returnpc);
  fflush(stdout);
#endif
  return frame->returnpc;
}


/*
 * å˿ե졼롣
 * : invokeJavaMethod() in javasrc/src/share/java/runtime/classruntime.c
 */
JavaFrame *allocNewJavaFrame(struct methodblock *mb, ExecEnv *ee) {
  int nlocals = (mb ? mb->nlocals : 0);
#if 0	/* ʬǵˡ */
  mb->nlocals = nlocals =
	(mb->fb.access & ACC_NATIVE)?0:
		(mb->args_size + ((mb->fb.access & ACC_STATIC)?0:1));
#endif

  return allocNewJavaFrameWithNLocals(mb, ee, nlocals);
}


/*
 * å˿ե졼롣
 * : invokeJavaMethod() in javasrc/src/share/java/runtime/classruntime.c
 */
JavaFrame *allocNewJavaFrameWithNLocals(struct methodblock *mb, ExecEnv *ee, int nlocals) {
  JavaFrame *prev_frame;
  stack_item *prev_optop;
  JavaStack *stack;
  JavaFrame *frame;

#ifdef DEBUG
    printf("allocNewJavaFrame((methodblock *)0x%x, (ExecEnv *)0x%x) is called.\n", (int)mb, (int)ee);
    if (mb)  printf(" method name: %s\n", mb->fb.name);
    fflush(stdout);
#endif

  if (!ee->current_frame) {	/* ҤȤĤ JavaFrame */
#ifdef DEBUG
    printf(" first frame.\n");  fflush(stdout);
#endif
    prev_frame = NULL;
#ifdef DEBUG
    if (!ee->initial_stack) {
      /* not reached */
      printf(" ee->initial_stack is NULL!\n");  fflush(stdout);
    }
#endif
    prev_optop = ee->initial_stack->data;
    stack = ee->initial_stack;
  }
  else {
    prev_frame = ee->current_frame;
    prev_optop = prev_frame->optop;
    stack = prev_frame->javastack;	/* current stack chunk */
  }
#ifdef DEBUG
  printf(" stack:       0x%x\n", (int)stack);
  if (stack)
    printf(" stack->next: 0x%x\n", (int)stack->next);
  fflush(stdout);
#endif

  frame = (JavaFrame *)(prev_optop + nlocals);

  if (frame->ostack + (mb ? mb->maxstack : 0) >= stack->end_data) {
    /* current stack chunk ˿ frame ޤʤ */
#ifdef DEBUG
    printf(" new frame doesn't fall into current stack chunk.\n");
    fflush(stdout);
#endif
    if (stack->next != 0) {
#ifdef DEBUG
      printf(" use next stack chunk.\n");
      fflush(stdout);
#endif
      stack = stack->next;
    }
    else {
      /*  stack chunk  */
#ifdef DEBUG
    printf(" must allocate new stack chunk.\n");
    fflush(stdout);
#endif
      if (stack->stack_so_far +
		JAVASTACK_CHUNK_SIZE*sizeof(stack->data[0]) > JavaStackSize) {
	SignalError(EE(), "java/lang/StackOverflowError", 0);
		/* λ ExecEnv 뤿ˡ ee ǤϤʤ EE() */
	return FALSE;
      }

      stack = CreateNewJavaStack(ee, stack);
      if (stack == 0) {
	SignalError(ee, "java/lang/OutOfMemoryError", 0);  return FALSE;
      }
    }

    /*  stack chunk ˡ frame ξ */
    frame = (JavaFrame *)(stack->data + nlocals);
    if (frame->ostack + (mb ? mb->maxstack : 0) >= stack->end_data) {
      /* stack chunk  frame ޤʤ */
      SignalError(EE(), "java/lang/InternalError", 0);  return FALSE;
    }

    prev_optop = stack->data;
  }	/* if (stack chunk  frame ޤʤ) */
#ifdef DEBUG
    printf(" frame: 0x%x\n", (int)frame);
    fflush(stdout);
#endif


  /* ¤ΥФ */
  frame->javastack = stack;
  frame->prev = prev_frame;
  frame->vars = prev_optop;
  frame->optop = frame->ostack;

  frame->current_method = mb;
  frame->constant_pool = (mb ? cbConstantPool(fieldclass(&mb->fb)) : NULL);
  frame->returnpc = (mb ? mb->code : 0);
  frame->monitor = 0;
  ee->current_frame = frame;

  frame->mon_starttime = 0;
  /* if (java_monitor) { frame->mon_starttime = now(); } */

#ifdef DEBUG
  printf(" constant_pool: 0x%x\n", (int)frame->constant_pool);

  printf("allocNewJavaFrame() done.\n");
  fflush(stdout);
#endif

  return frame;
}


/*
 * JavaFrame Ȥפ롣
 * for debug
 */
static void dumpAllFrames(FILE *out, JavaFrame *topFrame) {
  JavaFrame *frame;
  int count;

  fprintf(out, "Now dump all frames:\n");
  fprintf(out, "\n");

  /* stack frame  */
  frame = topFrame;
  count = 0;
  while (frame) {
    count++;
    frame = frame->prev;
  }

  frame = topFrame;
  while (frame) {
    fprintf(out, "dumping stack frame(%d):\n", --count);
    dumpFrame(out, frame);
    printf("\n");

    frame = frame->prev;
  }
}


static void dumpFrame(FILE *out, JavaFrame *frame) {
  struct methodblock *method = NULL;
  ClassClass *clazz = NULL;
  int isReturnpcZero = 0;	/* false */

  fprintf(out, "address of frame:\t0x%x\n", (int)frame);
  fflush(out);

  if (!frame)  return;

  method = frame->current_method;
  if (method)  clazz = method->fb.clazz;

  /* returnpc */
  if (!frame->returnpc)  isReturnpcZero = 1;	/* true */

  fprintf(out, "returnpc:\t\t");
  outPC(out, frame->returnpc, frame->current_method);  fprintf(out, "\n");
  fflush(out);

  /* constant_pool */
  /*  returnpc ˰¸
   *  returnpc is 0: (struct methodblock *)
   *       is not 0: (cp_itemm *)
   */
  {
    fprintf(out, "constant_pool:\t\t0x%x", (int)frame->constant_pool);
    fflush(out);

    if (!frame->constant_pool) {
      fprintf(out, " (0 is illegal!)\n");
    }
    else {
      if (isReturnpcZero) {
	struct methodblock *targetMethod =
	  (struct methodblock *)frame->constant_pool;
	fprintf(out, " (method: %s#%s)",
		cbName(targetMethod->fb.clazz), targetMethod->fb.name);
      }
      else {
	ClassClass *cpClazz = cp2class(frame->constant_pool);
	int index;
	if (cpClazz) {
	  /* CP is resolved. */
	  cp_item_type *cbCP = cbConstantPool(cpClazz);
	  index = (int)cp2offset(cpClazz, frame->constant_pool);
	  fprintf(out, " (offset: %d) in %s", index, cbName(cpClazz));
	}
	else {
	  index = frame->constant_pool->i;
	  fprintf(out, " (offset: %d)", index);
	}
      }
    }

    fprintf(out, "\n");
    fflush(out);
  }

  /* optop */
  {
    int optopOffset;
    optopOffset = (int)sp2offset(frame, frame->optop);
    fprintf(out, "optop:\t\t\t0x%x (offset: %d)\n",
	(int)frame->optop, optopOffset);
    fflush(out);
  }

  /* vars */
  {
    int varsOffset;
    varsOffset = (int)sp2offset(frame, frame->vars);
    fprintf(out, "vars:\t\t\t0x%x (offset: %d)\n",
	(int)frame->vars, varsOffset);
    fflush(out);
  }

  /* prev */
  fprintf(out, "prev:\t\t\t0x%x\n", (int)frame->prev);
  /* javastack */
  fprintf(out, "javastack:\t\t0x%x\n", (int)frame->javastack);

  /* lastpc */
  fprintf(out, "lastpc:\t\t\t");
  outPC(out, frame->lastpc, frame->current_method);  fprintf(out, "\n");
  fflush(out);

  /* current_method */
  fprintf(out, "current_method:\t\t0x%x (%s#%s)\n",
	  (int)method,
	  (method ? (cbName(method->fb.clazz)) : "null"),
	  (method ? (method->fb.name) : "null"));
  fflush(out);

  /* OTHERS */
  fprintf(out, "monitor:\t\t0x%x\n", (int)frame->monitor);
  fprintf(out, "mon_starttime:\t\t%d\n", frame->mon_starttime);
  fprintf(out, "ostack:\t\t\t0x%x\n", (int)frame->ostack);
  /* fprintf(out, ":\t0x%x\n", frame->); */

  fflush(out);

  return;
}
