/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdio.h>	/* for NULL */
#include <string.h>	/* for memcpy() */

#include <native.h>	/* for native methods */
#include <monitor.h>	/* for BINCLASS_{,UN}LOCK() */

#include "moba.h"
#include "opcodes.h"	/* for opc_* and opcode_length[] */


/*
 * Variables
 */
#define EXCEPTION_BUF_SIZE	100
static char exception_msg_buf[EXCEPTION_BUF_SIZE];


/*
 * PC б᥽åɤ롣
 * ɤƤ뤹٤ƤΥ饹õ <- 
 */
/* following code is derived from
	javasrc/src/share/java/runtime/classruntime.c */
extern struct methodblock *pc2method(unsigned char *pc) {
  struct methodblock *method;
  ClassClass *clazz;
  int i, j;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();
#if 0
  printf("nbinclasses: %d\n", clazzCount);  fflush(stdout);
#endif

#if (JDK_VER >= 110)
  BINCLASS_LOCK();
#endif /* JDK_VER >= 110 */
  for (i = clazzCount; i-- > 0; ) {
    clazz = clazzArray[i];
    method = cbMethods(clazz);
#if 0
    printf("class: %s\n", cbName(clazz));
    printf("nmethods: %d\n", cbMethodsCount(clazz));  fflush(stdout);
#endif
    for (j = cbMethodsCount(clazz); j-- > 0; method++) {
      if (!(method->fb.access & ACC_NATIVE)) {
	unsigned char *code_start;
	if (method->fb.access & ACC_MACHINE_COMPILED) {
	  if (PCinCompiledCode(pc, method))
	    goto ret;
	}
	else {
#if 0
	  printf(" method: %s starts 0x%x\n",
		 method->fb.name, (int)method->code);
	  fflush(stdout);
#endif
	  code_start = (unsigned char *)method->code;
	  if (pc >= code_start && pc < code_start + method->code_length)
	    goto ret;
	}
      }
    }
  }

  /* not found */
  return NULL;

ret:
#if (JDK_VER >= 110)
  BINCLASS_UNLOCK();
#endif /* JDK_VER >= 110 */
  return method;
}


/*
 * constant pool item б륯饹롣
 */
extern ClassClass *cp2class(cp_item_type *cp) {
  ClassClass *clazz;
  int i;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();

  for (i = clazzCount; i-- > 0; ) {
    cp_item_type *cbCP;
    cp_item_type *cbCPEnd;

    clazz = clazzArray[i];
    cbCP = cbConstantPool(clazz);
    cbCPEnd = cbCP + cbConstantPoolCount(clazz);

    if ((cp >= cbCP) && (cp < cbCPEnd))
      goto ret;
  }

  /* not found */
  return NULL;

ret:
  return clazz;
}


/*
 * FindClass()  wrapper
 * global reference ؤϿԤ
 */
extern ClassClass *FindAndRegisterClass(
	struct execenv *ee, char *name, bool_t resolve,
	struct Hjava_lang_ClassLoader *loader) {
  ClassClass *cb = NULL;
  if (loader != NULL)
    cb = ClassLoaderFindClass(ee, loader, name, resolve);
  else
    cb = FindClass(ee, name, resolve);

  /* Class ֥Ȥ GC ǲ뤳Ȥɤ */
  if (cb) {
    registerGlobalHandle((Hjava_lang_Object *)cbHandle(cb));
  }

  return cb;
}


/*
 * 饹λĥ᥽åɤõ
 */
extern struct methodblock *find_method_in_class(ExecEnv *ee, ClassClass *clazz,
				char *targetName, char *targetSig) {
  uint32_t targetId = (uint32_t)NameAndTypeToHash(targetName, targetSig);
  struct methodblock *method = cbMethods(clazz);
  int i;

  for (i = cbMethodsCount(clazz) - 1; i >= 0; i--) {
    char *name = (method + i)->fb.name;
    char *sig = (method + i)->fb.signature;
    uint32_t id = (uint32_t)NameAndTypeToHash(name, sig);
#if 0
    printf("find_m_i_class(%d): %s %s\n", i, name, sig);
    fflush(stdout);
#endif

    if (id == targetId) {
      method += i;
      break;
    }
  }
  if (i < 0) {	/* ᥽åɤĤʤ */
    jio_snprintf(exception_msg_buf, EXCEPTION_BUF_SIZE,
		"%s#%s sig.:%s", cbName(clazz), targetName, targetSig);
    SignalError(0, MOBA_PACKAGE "FieldNotFoundException",
		exception_msg_buf);
    return NULL;
  }

  return method;
}


/*
 * object(handle) λĥ᥽åɤõ
 */
extern struct methodblock *find_method_in_object(ExecEnv *ee, JHandle *h,
					char *targetName, char *targetSig) {
  uint32_t targetId = (uint32_t)NameAndTypeToHash(targetName, targetSig);
  struct methodtable *mtable = h->methods;
  ClassClass *clazz = mtable->classdescriptor;
  struct methodblock **methods = mtable->methods;
  int mtable_size = cbMethodTableSize(clazz);
  int i;

  for (i = mtable_size - 1; i >= 0; i--) {
    char *name = methods[i]->fb.name;
    char *sig = methods[i]->fb.signature;
    uint32_t id = (uint32_t)NameAndTypeToHash(name, sig);
#if 0
    printf("find_m_i_obj(%d): %s#%s %s\n",
		i, cbName(methods[i]->fb.clazz), name, sig);
    fflush(stdout);
#endif

    if (id == targetId) {
      break;
    }
  }
  if (i < 0) {	/* ᥽åɤĤʤ */
    jio_snprintf(exception_msg_buf, EXCEPTION_BUF_SIZE,
		"%s#%s sig.:%s", cbName(clazz), targetName, targetSig);
    SignalError(0, MOBA_PACKAGE "FieldNotFoundException",
		exception_msg_buf);
    return NULL;
  }

  return methods[i];
}


/*
 * NameAndTypeToHash()  ID  field õ
 * Ĥʤ NULL ֤
 */
struct fieldblock *findFieldByID(ClassClass *cb, uint32_t id) {
  struct fieldblock *field = cbFields(cb);
  int i;

  for (i = cbFieldsCount(cb); i-- > 0; field++) {
#if 1
    if (field->ID == id)
#else	/* ID 򤳤ξǷ׻ */
    if (NameAndTypeToHash(field->name, field->signature) == id)
#endif
      return field;
  }
  return NULL;
}

/*
 * NameAndTypeToHash()  ID  method õ
 * Ĥʤ NULL ֤
 */
struct methodblock *findMethodByID(ClassClass *cb, uint32_t id) {
  struct methodblock *method = cbMethods(cb);
  int i;

#if 0
  printf("findMethodByID(class `%s'):\n", cbName(cb));
#endif
  for (i = cbMethodsCount(cb); i-- > 0; method++) {
    struct fieldblock field = method->fb;
#if 0
    printf(" name: %s\n", field.name);
    printf(" sig : %s\n", field.signature);
#endif
#if 1
    if (field.ID == id)
#else	/* ID 򤳤ξǷ׻ */
    if (NameAndTypeToHash(field.name, field.signature) == id)
#endif
      return method;
  }
  return NULL;
}

/*
 * JavaVM ¸ߤ륯饹٤Ƥ field õ
 */
struct fieldblock *findFieldInVM(uint32_t ID) {
  struct fieldblock *field;
  ClassClass *clazz;
  int i;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();

  for (i = clazzCount; i-- > 0; ) {	/* ٤ƤΥ饹é */
    clazz = clazzArray[i];

    field = findFieldByID(clazz, ID);
    if (field)  goto returnFindFieldInVM;
  }
  field = NULL;	/* not found */

returnFindFieldInVM:
  return field;
}

/*
 * JavaVM ¸ߤ륯饹٤Ƥ method õ
 */
struct methodblock *findMethodInVM(uint32_t ID) {
  struct methodblock *method;
  ClassClass *clazz;
  int i;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();

#if (JDK_VER >= 110)
  BINCLASS_LOCK();
#endif /* JDK_VER >= 110 */
  for (i = clazzCount; i-- > 0; ) {	/* ٤ƤΥ饹é */
    clazz = clazzArray[i];

    method = findMethodByID(clazz, ID);
    if (method)  goto returnFindMethodInVM;
  }
  method = NULL;	/* not found */

returnFindMethodInVM:
#if (JDK_VER >= 110)
  BINCLASS_UNLOCK();
#endif /* JDK_VER >= 110 */
  return method;
}



/*
 * ɽѿ̾롣
 */
char *LocalVarName(struct methodblock *method, int index) {
  cp_item_type *cpool = cbConstantPool(method->fb.clazz);
  struct localvar *lv = method->localvar_table + index;

  return cpool[lv->nameoff].cp;
}

/*
 * ɽѿ signature 롣
 */
char *LocalVarSig(struct methodblock *method, int index) {
  cp_item_type *cpool = cbConstantPool(method->fb.clazz);
  struct localvar *lv = method->localvar_table + index;

  return cpool[lv->sigoff].cp;
}

/*
 * Ϳ줿 PC λǶɽѿƤ뤫?
 */
int isLocalVarAlive(struct methodblock *method, int index, uint32_t pcoff) {
  cp_item_type *cpool = cbConstantPool(method->fb.clazz);
  struct localvar *lv = method->localvar_table + index;
  long start, len, end;

  start = lv->pc0;  len = lv->length;  end = start + len;

  if ((start < 0) || (len < 0))  return 0;
  
  if ((pcoff >= start) && (pcoff < end))
    return 1;
  else
    return 0;
}


/*
 *  opcode  PC 롣
 */
unsigned char *nextPC(unsigned char *pc) {
  int size_mag = 1;
  int opcode = (int)*pc;
  int opcodeLen = (int)opcode_length[opcode];
  /*
    dumpOpcodeLength(stdout);
    fflush(stdout);
   */
#ifdef DEBUG
  printf("opcode: %s(0x%x=%d), length of opcode: %d  (in nextPC())\n",
		opcode_symbol[opcode], opcode, opcode, opcodeLen);
  printf("PC: 0x%x ->", (int)pc);
#endif

  if (opcode == opc_wide) {
    opcode = *(pc+1);
    size_mag = 2;
  }

  pc += opcodeLen * size_mag;
#ifdef DEBUG
  printf(" 0x%x.\n", (int)pc);
  fflush(stdout);
#endif

  return pc;
}

/*
 * unsigned char opcode_length[]  dump 롣(for debug)
 */
void dumpOpcodeLength(FILE *out) {
  int i;
  for (i = 0; i <= opc_nonnull_quick; i++) {
    fprintf(out, "len[%d]: %d\n", i, (int)opcode_length[i]);
  }
}


/*
 * ݥ󥿤λؤ褬 handle ɤȽꤹ롣
 */
extern int isPtrHandle(void *p) {
/* ( #include <gc.h>)
  SetLimits();  return ValidHandle(p);
 */
/* ȤɻȤʤ !!! */

struct methodtable *methods;
ClassClass *clazz;

#ifdef DEBUG
  printf("isPtrHandle(0x%x) called.\n", (int)p);
  fflush(stdout);
#endif
  if ( !validateAddress(p) )  return 0;	/* false */

  if ( !validateAddress((void *)unhand((JHandle *)p)) )  return 0;

  if (obj_flags((JHandle *)p) == T_NORMAL_OBJECT) {	/* ֥? */
    if ( !validateAddress((void *)(((JHandle *)p)->methods)) )  return 0;
    methods = (struct methodtable *)((JHandle *)p)->methods;

    if ( !validateAddress((void *)(methods->classdescriptor)) )  return 0;
    clazz = (ClassClass *)methods->classdescriptor;
    if ( !validateAddress((void *)cbName(clazz)) )  return 0;
    if ( !validateAddress((void *)cbSuperName(clazz)) )  return 0;
    if ( !validateAddress((void *)cbSourceName(clazz)) )  return 0;
  }
  else {	/* ? */
    int t = (int)p;
    if (!(t >= T_CLASS && t < T_MAXNUMERIC))  return 0;
  }

#if 0
  printf("class name: %s\n", cbName(clazz));
  fflush(stdout);
#endif
  

#ifdef DEBUG
  printf("isPtrHandle(0x%x) returns true.\n", (int)p);
  fflush(stdout);
#endif
  return 1;	/* true */
}


/*
 * char 󤫤 Byte[] 롣
 */
extern HArrayOfByte *makeArrayOfByte(char *str, int len) {
  HArrayOfByte *ret;
  char *body;

#ifdef DEBUG
  printf("makeArrayOfByte(str:0x%x, len:%d) called.\n", (int)str, len);
  fflush(stdout);
#endif
  ret = (HArrayOfByte *) ArrayAlloc(T_BYTE, len);
  if (!ret) {
    SignalError(0, "java/lang/OutOfMemoryError", 0);
    return 0;
  }

  body = ((ArrayOfByte *)unhand(ret))->body;
  if (str)
    memcpy(body, str, len);
  else
    memset(body, 0, len);
#if (JDK_VER >= 110)
  KEEP_POINTER_ALIVE(body);
#endif /* JDK_VER >= 110 */
  return ret;
}


/*
 * (unicode *)ʸϤ롣
 * This function is not thread safe.
 */
#define UNICODE_BUF_LEN	50
extern void outUnicode(FILE *out, unicode *str) {
  static char buf[UNICODE_BUF_LEN];

  unicode2str(str, buf, UNICODE_BUF_LEN);
  fprintf(out, "%s", buf);

  fflush(out);
}


/*
 * PCϤ롣
 */
extern void outPC(FILE *out, unsigned char *pc, struct methodblock *method) {
  int offset;

  if (pc == 0) {
    fprintf(out, "0x0");  return;
  }

  if (!method)
    method = pc2method(pc);

  fprintf(out, "0x%x", pc);
  if (method) {
    offset = (int)pc2offset(method, pc);
    fprintf(out, " (offset: %d)", offset);
  }
  fprintf(out, " in ");
  outMethod(out, method);
}


/*
 * ᥽åɤˤĤƽϤ롣
 */
extern void outMethod(FILE *out, struct methodblock *method) {
  ClassClass *clazz;

  if (!method) {
    fprintf(out, "(null)#(null)");  fflush(out);
    return;
  }

  clazz = method->fb.clazz;

  fprintf(out, "%s#%s", (clazz ? cbName(clazz): "(null)"), method->fb.name);
  fflush(out);
}


/*
 * PC 򥳡ƬΥեåȤѴ롣
 * եåȤñ̤ sizeof(unsigned char)
 */
extern uint32_t pc2offset(struct methodblock *method, unsigned char *pc) {
  uint32_t ret;

  if (!pc)  ret = -1;
  else  ret = (uint32_t)(pc - method->code);

  return  ret;
}

/*
 * ƬΥեåȤ PC Ѵ롣
 */
extern unsigned char *offset2pc(struct methodblock *method, uint32_t offset) {
  if (offset == -1)
    return 0;
  else
    return (method->code + offset);
}

/*
 * åݥ󥿤򥹥å줫ΥեåȤѴ롣
 * եåȤñ̤ sizeof(stack_item)
 */
extern int16_t sp2offset(JavaFrame *frame, stack_item *sp) {
  int16_t ret;

  if (!sp)  ret = -1;
  else  ret = (int16_t)(sp - frame->ostack);

  return ret;
}

/*
 * å줫ΥեåȤ򥹥åݥ󥿤Ѵ롣
 */
extern stack_item *offset2sp(JavaFrame *frame, int16_t offset) {
  if (offset == -1)
    return 0;
  else
    return (frame->ostack + offset);
}

/*
 * 󥹥ȥס(CP)ǤؤΥݥ󥿤򥪥եåȤѴ롣
 */
extern uint16_t cp2offset(ClassClass *clazz, cp_item_type *cp_item) {
  uint16_t ret;

  if (!cp_item)  ret = -1;
  else  ret = (uint16_t)(cp_item - cbConstantPool(clazz));

  return ret;
}

/*
 * եåȤ CP ǤؤΥݥ󥿤Ѵ롣
 */
extern cp_item_type *offset2cp(ClassClass *clazz, uint16_t offset) {
  if ((int16_t)offset == -1)
    return 0;
  else
    return (cbConstantPool(clazz) + offset);
}


/*
 * invoke* ̿᤬ƤӽФ᥽åɤ롣
 * ʤ NULL ֤
 */
extern struct methodblock *calleeByInvokeOpcode(
			struct methodblock *method, unsigned char *lastpc) {
  ExecEnv *ee = EE();
  ClassClass *clazz;
  cp_item_type *constant_pool;
  int CPindex;
  struct methodblock *result = NULL;

  unsigned char *typeTable;
  int resolved;

  if (!method)  return NULL;

  clazz = method->fb.clazz;

  switch (*lastpc) {
  case opc_invokevirtual:
  case opc_invokespecial:
  case opc_invokestatic:
  case opc_invokeinterface:
  case opc_invokenonvirtual_quick:
  case opc_invokevirtual_quick_w:
  case opc_invokestatic_quick:
  case opc_invokeinterface_quick:
    constant_pool = cbConstantPool(clazz);
    CPindex = ((int)*(lastpc+1) << 8) | *(lastpc+2);
    typeTable =
	(unsigned char *)constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
    resolved = typeTable[CPindex] & CONSTANT_POOL_ENTRY_RESOLVED;
    if (!resolved)
      ResolveClassConstantFromClass(clazz, (unsigned)CPindex, ee,
					1 << CONSTANT_Methodref);
    result = constant_pool[CPindex].mb;
    break;
  case opc_invokesuper_quick:
    clazz = cbSuperclass(clazz);
    constant_pool = cbConstantPool(clazz);
    CPindex = ((int)*(lastpc+1) << 8) | *(lastpc+2);
    typeTable =
	(unsigned char *)constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].cp;
    resolved = typeTable[CPindex] & CONSTANT_POOL_ENTRY_RESOLVED;
    if (!resolved)
      ResolveClassConstantFromClass(clazz, (unsigned)CPindex, ee,
					1 << CONSTANT_Methodref);
    result = constant_pool[CPindex].mb;
    break;

  case opc_invokevirtual_quick:
  case opc_invokevirtualobject_quick:
  case opc_invokeignored_quick:
#ifdef DEBUG
    printf("In calleeByInvokeOpcode():\n");
    printf("  don't support this opcode: %d\n", (int)*lastpc);
    fflush(stdout);
#endif
    break;

  default:
#ifdef DEBUG
    printf("In calleeByInvokeOpcode():\n");
    printf("  invalid opcode: %d\n", (int)*lastpc);
    fflush(stdout);
#endif
    break;
  }

  return result;
}



/*
 * signature Ƭ򸫤ơtypecode (T_*) ֤
 */
int signature2typecode(char *sig) {
  int type;
  switch (*sig) {
  case SIGNATURE_ARRAY:  type = T_CLASS; break;
	/*  T_ARRAY ¸ߤʤ */
  case SIGNATURE_BYTE:   type = T_BYTE; break;
  case SIGNATURE_CHAR:   type = T_CHAR; break;
  case SIGNATURE_CLASS:  type = T_CLASS; break;
  /*case SIGNATURE_ENUM:*/
  case SIGNATURE_FLOAT:  type = T_FLOAT; break;
  case SIGNATURE_DOUBLE: type = T_DOUBLE; break;
  case SIGNATURE_INT:    type = T_INT; break;
  case SIGNATURE_LONG:   type = T_LONG; break;
  case SIGNATURE_SHORT:  type = T_SHORT; break;
  case SIGNATURE_VOID:   type = T_VOID; break;
  case SIGNATURE_BOOLEAN:type = T_BOOLEAN; break; 
  defaut:		 type = 0; break;
  }
  return type;
}


/*
 * htonl, htons, ntohl, ntohs  64bit 
 */
#ifndef BIG_ENDIAN
extern uint64_t hton64(uint64_t x) {
  uint64_t ret;

  *(unsigned long *)&ret = htonl((unsigned long)(x >> 32));
  *(((unsigned long *)&ret) + 1) = htonl((unsigned long)x);

  return ret;
}

extern uint64_t ntoh64(uint64_t x) {
  uint64_t ret = 0;

  unsigned long l1 = ( (uint64_t)ntohl(*(unsigned long *)&x) );
  unsigned long l2 = (uint64_t)ntohl(*(((unsigned long *)&x) + 1));

  ret |= ( (uint64_t)ntohl(*(unsigned long *)&x) ) << 32;
  ret |= (uint64_t)ntohl(*(((unsigned long *)&x) + 1))  & 0x00000000ffffffffLL;

  return ret;
}
#endif	/* BIG_ENDIAN */
