/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.place;


import java.net.Socket;
import java.net.InetAddress;
import java.io.InputStream;		// for negotiateClassDef()
import java.io.OutputStream;
import java.io.DataInputStream;		// for negotiateClassDef()
import java.io.BufferedInputStream;	// for optimization of throughput
import java.io.BufferedOutputStream;	// for optimization of throughput
import java.io.FileOutputStream;	// for debug
import java.io.IOException;
import NET.shudo.moba.lang.MobaThread;
import NET.shudo.moba.lang.PlaceAddress;
import NET.shudo.moba.io.ClassWriter;
import NET.shudo.moba.io.ThreadWriter;
import NET.shudo.moba.reflect.MetaClass;// for negotiateClassDef()


/**
 * åɤФ饹<BR>
 * ľܤ MobaThread#goTo(), MobaThread#moveTo(),
 * MobaThread#fork(), MobaThread#forkThisThread() Ȥ٤
 *
 * @see NET.shudo.moba.lang.MobaThread#goTo
 * @see NET.shudo.moba.lang.MobaThread#fork
 */
public
class ThreadExporter {
  /**
   * Ϳ줿åɤ¾η׻˰ư롣<BR>
   * ľܤϡMobaThread.moveTo(), goTo() Ѥ٤
   *
   * @see NET.shudo.moba.lang.MobaThread#goTo()
   * @see NET.shudo.moba.lang.MobaThread#moveTo()
   */
  public static void moveTo(MobaThread thread, PlaceAddress addr)
	throws IOException {
    try {
      ThreadExporter.fork(thread, addr);
    }
    catch (IOException e) {
      // fork failed.
      // åɤϻߤʤ
      throw e;
    }

    if (!thread.isInternalized()) {
      // ưΥåɤߤ
      try {
	thread.stop();
      }
      catch (SecurityException e) {
	// not reached.
	e.printStackTrace();  System.exit(1);
      }
    }

    Logger.log(null, "sending a thread succeed");
  }


  /**
   * Ϳ줿åɤ¾η׻˥ԡ롣(cf. fork(2))
   * ľܤϡMobaThread.moveTo(), goTo() Ѥ٤
   *
   * @see NET.shudo.moba.lang.MobaThread#goTo()
   * @see NET.shudo.moba.lang.MobaThread#moveTo()
   *
   * @return true to the origianl thread,
   * false to the thread created on remote machine.
   */
  public static boolean fork(MobaThread thread, PlaceAddress addr)
	throws IOException {
    Thread crntThread = Thread.currentThread();

//System.out.println("fork(" + thread + ", " + addr + ") called.");
    if (!thread.equals(crntThread)) {
//System.out.println("move otherself.");
      try {
	// ɬפǤϤʤ suspend Ƥ
        thread.suspend();
      }
      catch (SecurityException e) {
	throw new IOException("Security Exception caused.");
      }
    }
    else {
//System.out.println("move myself.");
    }


//try{
//(new ThreadWriter(new FileOutputStream("output"))).writeObject(thread);
//}
//catch (Exception e) { e.printStackTrace(); System.exit(1); }


    int threadPort = addr.port() + PlaceAddress.THREAD_PORT_DIFF;
    Socket sock = null;
    sock = new Socket(addr.inetAddress(), threadPort);	// throws IOException

    OutputStream sockOut = null;
    InputStream sockIn = null;
//    sockOut = new BufferedOutputStream(sock.getOutputStream(), 32 * 1024);
//    sockIn = new BufferedInputStream(sock.getInputStream(), 32 * 1024);
    sockOut = sock.getOutputStream();
    sockIn = sock.getInputStream();

/* obsolete
    // 饹ΤȤ
    negotiateClassDef(thread, sockOut, sockIn);
*/

    ThreadWriter out = new ThreadWriter(sockOut);

    // åɤ줿Ȥ
    out.writeObject(thread.generatedPlace());


    // åɤν񤭽Ф
    out.writeObject(thread);		// throws IOException
    out.flush();


    boolean ret = !thread.isInternalized();
	// ΥåɤǤϿˤʤ
    if (ret) {
      // close
      sock.close();
    }

    return ret;
  }


  /**
   * ThreadImporter Ȥδ֤ɬפʥ饹ΤȤԤ
   */
  private static void negotiateClassDef(MobaThread thread,
	OutputStream sockOut, InputStream sockIn) throws IOException {
    ClassWriter out = new ClassWriter(sockOut);
    DataInputStream in = new DataInputStream(sockIn);

    String[] classes;
    int nClass;
    int i;

    // 饹̾
    Runnable thrTarget = thread.runnableInstance();
    if (thrTarget != null)
      classes = (MetaClass.get(thrTarget)).referringClassnames();
    else
      classes = (MetaClass.get(thread)).referringClassnames();

    out.writeInt(classes.length);
    for (i = 0; i < classes.length; i++) {
      out.writeUTF(classes[i]);
    }
    out.flush();

    // 饹̾
    nClass = in.readInt();
    classes = new String[nClass];
    for (i = 0; i < nClass; i++) {
      classes[i] = in.readUTF();
    }

    // 饹
    for (i = 0; i < nClass; i++) {
      MetaClass mc = null;
      try {
	mc = MetaClass.get(classes[i]);
      }
      catch (ClassNotFoundException e) {}

      if (mc != null) {
	out.writeInt((int)mc.classfileLength());
	out.writeClass(mc.classClass());
      }
      else {
	out.writeInt(0);
      }
    }
    out.flush();
  }
}
