/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.reflect;

import java.io.OutputStream;	// for write*InClassfile()
import java.io.IOException;	// for write,readObject()
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Enumeration;	// for classNamesInSignature()
import java.util.Vector;	// for referringClassnames()
import java.util.Stack;		// for referringClassnames()
import NET.shudo.moba.lang.GlobalReference;
import NET.shudo.moba.io.ObjectWriter;
import NET.shudo.moba.io.ObjectReader;

public final
class MetaClass {
  private static Dictionary metaClassTable = new Hashtable();
	// key  Class  instance

  private static final String STREAMABLE_CLASS_NAME =
	"NET.shudo.moba.io.Streamable";
  private static Class streamableClass = null;
  private static Object streamableClassLock = new Object();


  private transient MetaField[] fields = null;
  private transient MetaMethod[] methods = null;

  private transient Class thisClass = null;
  private transient Dictionary fieldNameTable = null;
  private transient Dictionary methodSelectorTable = null;

  private transient boolean hasObjRead;
  private transient boolean hasObjWrite;

  private long classfileLen = -1;


  public MetaClass() {}

  /**
   *  MetaClass  Class ǽ롣
   */
  private void init(Class thisClass) {
    int nFields = 0;
    int nMethods = 0;

    this.thisClass = thisClass;

    // MetaField 롣
    nFields = fieldCount();
    this.fields = new MetaField[nFields];
    for (int i = 0; i < nFields; i++) {
      this.fields[i] = this.createField(i);
    }
    // MetaMethod 롣
    nMethods = methodCount();
    this.methods = new MetaMethod[nMethods];
    for (int i = 0; i < nMethods; i++) {
      this.methods[i] = this.createMethod(i);
    }

    // field, method name ɽ롣̾ -> index
    fieldNameTable = new Hashtable();
    for (int i = 0; i < fieldCount(); i++) {
      fieldNameTable.put(fields[i].name(), new Integer(i));
    }
    methodSelectorTable = new Hashtable();
    for (int i = 0; i < methodCount(); i++) {
      MetaMethod method = methods[i];
      methodSelectorTable.put(new Integer(method.ID()),
				new Integer(i));
    }

    // writeObject(), readObject() äƤ뤫Ĵ٤롣
    this.hasObjWrite = hasOwnWriteObject0();
    this.hasObjRead = hasOwnReadObject0();

    // MetaClass ɽϿ (ѤΤ)
    synchronized(metaClassTable) {
      metaClassTable.put(this.thisClass, this);
    }
  }

  // MetaClass Ȥϥ󥹥ȥ饯ǤϤʤƤ
  public static MetaClass get(Class thisClass) {
    // native methods 饤֥Υ
    GlobalReference.loadLibrary();

    MetaClass ret = null;

    synchronized(metaClassTable) {
      ret = (MetaClass)metaClassTable.get(thisClass);
      if (ret == null) {
	ret = new MetaClass();
	ret.init(thisClass);
      }
    }

    return ret;
  }
  public static MetaClass get(String className, ClassLoader loader)
	throws ClassNotFoundException {
    // native methods 饤֥Υ
    GlobalReference.loadLibrary();

    if (Signature.isArray(className))
      return getBySignature(className, loader);

    if (loader != null) {	// 饹ǥ
      Class clazz = loader.loadClass(className);
      return get(clazz);
    }
    else			// ƥˡǥ
      return get(Class.forName(className));
  }
  public static MetaClass get(String className) throws ClassNotFoundException {
    return get(className, null);
  }
  public static MetaClass get(Object obj) {
    if (obj == null)  return null;
    return get(obj.getClass());
  }
  public static MetaClass getBySignature(String signature, ClassLoader loader)
	throws ClassNotFoundException {
    // native methods 饤֥Υ
    GlobalReference.loadLibrary();

    return get(classBySignature(signature, loader));
  }
  public static MetaClass getBySignature(String signature)
	throws ClassNotFoundException {
    return getBySignature(signature, null);
  }

  public String toString() {
    if (thisClass == null) {
      // not reached.
      return "((MetaClass)this is not initialized)";
    }

    StringBuffer buf = new StringBuffer();
    int count;

    buf.append(thisClass.toString());
    count = fieldCount();
    if (count > 0) {
      buf.append('{');
      try {
	buf.append(field(0).name());
	for (int i = 1; i < count; i++) {
	  buf.append(',');
	  buf.append(field(i).name());
	}
      }
      catch (FieldNotFoundException e) {
	// not reached
	e.printStackTrace();  System.exit(1);
      }
      buf.append('}');
    }

    return buf.toString();
  }

  /**
   * java.lang.Class֤
   */
  public Class classClass() { return thisClass; }

  /**
   * 饹֤̾
   */
  public String getName() { return thisClass.getName(); }

  /**
   * signature ֤
   */
  public native String signature();

  /**
   * signature б륯饹롣
   */
  protected static native Class classBySignature(
	String signature, ClassLoader loader) throws ClassNotFoundException;


  /**
   * Υ饹Υ󥹥󥹤롣<BR>
   * κˤ newInstance(int) Ȥ
   */
  public Object newInstance() {
    Object ret = null;
    if (isArray())
      ret = null;
    else
      ret = newInstance(thisClass);
//System.out.println("MetaClass#newInstance() returns: " + ret);
    return ret;
  }
  /**
   *  () 饹Υ󥹥󥹤롣
   */
  public Object newInstance(int len) {
    Object ret = null;
    if (isArray()) {
      ret = newArray(thisClass, len);
    }
    else
      ret = newInstance(thisClass);
    return ret;
  }
  private static native Object newInstance(Class thisClass);
  private static native Object newArray(Class thisClass, int len);


  /**
   * 饹Ϳ줿ǥåĥեɤ롣
   */
  public MetaField field(int i) throws FieldNotFoundException {
    if (i < 0 || i >= this.fields.length) {
      throw new FieldNotFoundException("No such field: " + i);
    }
    return this.fields[i];
  }

  /**
   * 饹Ϳ줿ǥåĥ᥽åɤ롣
   */
  public MetaMethod method(int i) throws FieldNotFoundException {
    if (i < 0 || i >= this.methods.length) {
      throw new FieldNotFoundException("No such method: " + i);
    }
    return this.methods[i];
  }

  /**
   * Ϳ줿ե̾ĥեɤ롣
   */
  public MetaField field(String name) throws FieldNotFoundException {
    Integer index = (Integer)fieldNameTable.get(name);
    if (index == null) {
      throw new FieldNotFoundException("Field not found: " + name);
    }
    return this.fields[index.intValue()];
  }

  /**
   * Ϳ줿쥯 (᥽å̾signature) ĥ᥽åɤ롣
   */
  public MetaMethod method(String name, String sig)
	throws FieldNotFoundException {
    sig = sig.replace('.', '/');

    int ID = MetaMethod.methodID(name, sig);
    Integer index = (Integer)methodSelectorTable.get(new Integer(ID));
    if (index == null) {
      throw new FieldNotFoundException("Method not found: " + name);
    }
    return this.methods[index.intValue()];
  }

  /**
   * Ϳ줿ǥåΥեɥ֥Ȥ롣
   */
  private MetaField createField(int i) {
    MetaField aField = new MetaField(this, i);
    return aField;
  }

  /**
   * Ϳ줿ǥåΥ᥽åɥ֥Ȥ롣
   */
  private MetaMethod createMethod(int i) {
    MetaMethod aMethod = new MetaMethod(this.thisClass, i);
    return aMethod;
  }


  /**
   * Υ饹ʤ鿿֤<BR>
   * JDK 1.1 ʹߤ java.lang.Class#isArray() ƱΥ᥽åɡ
   */
  public boolean isArray() {
    //return thisClass.getName().startsWith("[");
    return Signature.isArray(signature());
  }

  /**
   * ֥ (or ) Υ饹ʤ鿿֤
   */
  public boolean isObjectArray() {
    return Signature.isObjectArray(signature());
  }

  /**
   * ܷ (byte, char, int, ...) ʤ鿿֤
   */
  public boolean isBaseTypeArray() {
    return Signature.isBaseTypeArray(signature());
  }


  /**
   * 饹롣<BR>
   * Υ᥽åɤϾʤͽꡣ
   */
  public static native Class defineClass(
	ClassLoader loader, String name, byte[] data, int offset, int length);

  /**
   * Υ饹ɬפȤ륯饹֤
   */
  public String[] referringClassnames() {
    Class[] arg = new Class[1];
    arg[0] = thisClass;

    return MetaClass.referringClassnames(arg);
  }
  /**
   * Υ饹ɬפȤ륯饹
   * 󥹥ȥס뤫Ū֤̾<BR>
   * java.*, sun.* Ͻsuperclass ˡ
   */
  public static String[] referringClassnames(Class[] clazzes) {
    int i, j;
    String[] classnames = null;
	// ̤ unify Ƥ
    Vector ret = new Vector();
    Stack classStack = new Stack();

    for (i = 0; i < clazzes.length; i++) {
      Class targetClazz = clazzes[i];
      if (targetClazz == null)  continue;
      classnames = ConstantPool.classnames(targetClazz);

      for (j = 0; j < classnames.length; j++) {
	String classname = classnames[j];
	Class clazz = null;

	if (classname.startsWith("java.") || classname.startsWith("sun."))
	  continue;	// ɸ९饹Ϥͤ

	try {
	  clazz = Class.forName(classname);
	}
	catch (ClassNotFoundException e) { continue; }

	// superclass -> subclass é
	do {
	  classStack.push(clazz);
	  clazz = clazz.getSuperclass();
	} while (clazz != null);
	while (!classStack.empty()) {
	  clazz = (Class)classStack.pop();
//System.out.println("{" + clazz.getName() + "}");  System.out.flush();
	  classname = clazz.getName();
	  if (classname.startsWith("java.") || classname.startsWith("sun."))
	    continue;	// ɸ९饹Ϥͤ
	  if (!ret.contains(clazz)) {
//System.out.println("add {" + clazz.getName() + "}");  System.out.flush();
	    ret.addElement(clazz);
	  }
	}
      }	// constant pool Υ饹é
    }	// Υ饹é

    classnames = new String[ret.size()];
    i = 0;
    for (Enumeration e = ret.elements(); e.hasMoreElements(); ) {
      classnames[i++] = ((Class)e.nextElement()).getName();
    }

    return classnames;
  }


  /**
   * writeObject()äƤ뤫
   */
  public boolean hasOwnWriteObject() { return this.hasObjWrite; }
  private native boolean hasOwnWriteObject0();

  /**
   * readObject()äƤ뤫
   */
  public boolean hasOwnReadObject() { return this.hasObjRead; }
  private native boolean hasOwnReadObject0();


  /**
   * implementsƤ뤫Ĵ٤롣
   */
  public boolean doesImplement(Class iface) {
    // JDK 1.1 Ǥ Class#isAssignableFrom() ǼǤ롣

    if (iface == null)  return false;

    String ifaceName = iface.getName();
    Class[] ifarray = thisClass.getInterfaces();

    for (int i = 0; i < ifarray.length; i++) {
      if (ifaceName.equals(ifarray[i].getName()))  return true;
    }
    return false;
  }

  /**
   * StreamableimplementsƤ뤫Ĵ٤롣
   */
  public boolean isStreamable() {
    synchronized(streamableClassLock) {
      if (streamableClass == null) {
	try {
	  streamableClass = Class.forName(STREAMABLE_CLASS_NAME);
	}
	catch (ClassNotFoundException e) {
	  // not reached.
	  e.printStackTrace();  System.exit(1);
	}
      }
    }

    return doesImplement(streamableClass);
  }


  /**
   * super class ֤
   */
  public MetaClass superClass() {
    Class superClass = thisClass.getSuperclass();
    if (superClass == null) {
      return null;
    }
    else {
      return MetaClass.get(superClass);
    }
  }


  /**
   * magic number ֤
   */
  public static native int magicNumber();

  /**
   * major version ֤(uint16)
   */
  public native int majorVersion();

  /**
   * minor version ֤(uint16)
   */
  public native int minorVersion();

  /**
   * constantpool count ֤(uint16)
   */
  public int constantpoolCount() {
    return ConstantPool.count(thisClass);
  }

  /**
   * access ե饰֤(uint16)
   */
  public native int access();

  /**
   * եɿ֤(uint16)
   */
  public native int fieldCount();

  /**
   * ᥽åɿ֤(uint16)
   */
  public native int methodCount();

  /**
   * interface count ֤(uint16)
   */
  public native int interfaceCount();


  /**
   * native methods ο֤
   */
  public int nativeMethodCount() {
    return methodCountWithAccess(Access.NATIVE);
  }

  /**
   * static methods ο֤
   */
  public int staticMethodCount() {
    return methodCountWithAccess(Access.STATIC);
  }

  /**
   * ꤷ access modifier ĥ᥽åɤο֤
   */
  private native int methodCountWithAccess(int mod);


  /**
   * classfile ΥեޥåȤ interface Ϥ롣
   */
  public native void writeIFInClassfile(OutputStream out) throws IOException;

  /**
   * ե֤̾
   */
  public native String sourceName();


  /**
   * ClassWriter 륯饹եĹ֤
   */
  public long classfileLength() {
    // attrib. ̾ index
    int utfConstantValueIndex =
	ConstantPool.utf8Index(thisClass, "ConstantValue");
    int utfCodeIndex = ConstantPool.utf8Index(thisClass, "Code");
    int utfExceptionsIndex = ConstantPool.utf8Index(thisClass, "Exceptions");
    int utfLineNoTblIndex =
	ConstantPool.utf8Index(thisClass, "LineNumberTable");
    int utfLVarTblIndex =
	ConstantPool.utf8Index(thisClass, "LocalVariableTable");
    int utfSourceFileIndex = ConstantPool.utf8Index(thisClass, "SourceFile");


    MetaClass mc = MetaClass.get(thisClass);
    int i, j;

    // 饹γƼ
    int cpCount = mc.constantpoolCount();
    int intfCount = mc.interfaceCount();
    int fieldCount = mc.fieldCount();
    int methodCount = mc.methodCount();
    String sourceName = mc.sourceName();

    // 줾 attrib. ¸ߤ뤫ʤ (ѿ *Exists)
    boolean sourceFileExists =
	((sourceName != null) && (utfSourceFileIndex > 0));


    if (classfileLen >= 0)  return classfileLen;


    // file Ƭ
    classfileLen = 10;
    classfileLen += ConstantPool.lengthInClassfile(thisClass);
    classfileLen += 8;
    classfileLen += 2 * intfCount;


    // field count
    classfileLen += 2;
    // field
    for (i = 0; i < fieldCount; i++) {
      MetaField mf = null;
      int access = 0;
      try {
	mf = mc.field(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mf.access();

      classfileLen += 8;
      if (Access.isStatic(access) && Access.isValknown(access)) {
	if (utfConstantValueIndex >= 0)
	  classfileLen += 8;
      }
    }	// եɤ


    // method count
    classfileLen += 2;
    // method
    for (i = 0; i < methodCount; i++) {
      MetaMethod mm = null;
      int access = 0;
      try {
	mm = mc.method(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mm.access();

      classfileLen += 8;

      // attrib. `Code'
      if (!Access.isNative(access)) {
	classfileLen += 14;
	classfileLen += mm.codeLength();

	int catchFrameLength = mm.catchFrameLength();
	int lineNoTableLength = mm.lineNoTableLength();
	int localVarTableLength = mm.localVarTableLength();
	boolean lineNoTableExists =
		((lineNoTableLength > 0) && (utfLineNoTblIndex > 0));
	boolean localVarTableExists =
		((localVarTableLength > 0) && (utfLVarTblIndex > 0));

	// catch frame
	classfileLen += 2;
	classfileLen += 8 * catchFrameLength;

	// attrib. in attrib. `Code'
	classfileLen += 2;
	// attrib. `LineNumberTable'
	if (lineNoTableExists) {
	  classfileLen += 8 + 4 * lineNoTableLength;
	}
	// attrib. `LocalVariableTable'
	if (localVarTableExists) {
	  classfileLen += 8 + 10 * localVarTableLength;
	}
      }	// attrib. `Code'

      // attrib. `Exceptions'
      int thrownExLength = mm.thrownExceptionLength();
      boolean thrownExExists =
		((thrownExLength > 0) && (utfExceptionsIndex > 0));
      if (thrownExExists) {
	classfileLen += 8 + 2 * thrownExLength;
      }
    }	// ᥽åɤé

    classfileLen += 2;

    // attrib. `Source'
    if (sourceFileExists)  classfileLen += 8;

    return classfileLen;
  }


  /**
   *  writeObject()
   */
  private void writeObject(ObjectWriter out) throws IOException {
//System.out.println("MetaClass#writeObject() called.");
//System.out.flush();
    out.writeUTF(this.signature());
  }

  /**
   *  readObject()
   */
  private void readObject(ObjectReader in)
	throws IOException, ClassNotFoundException {
//System.out.println("MetaClass#readObject() called.");
//System.out.flush();
    this.init(MetaClass.classBySignature(in.readUTF(), in.classLoader()));
//System.out.println(" class: " + this.getName());
//System.out.flush();
  }
}
